#! /usr/bin/env python
"""
Unit tests for landlab.components.flexure.flexure
"""
import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.components import Flexure

(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e3, 10e3), (0., 0.))


def test_method_names():
    grid = RasterModelGrid((20, 20), xy_spacing=10e3)
    assert Flexure(grid, method="airy").method == "airy"
    assert Flexure(grid, method="flexure").method == "flexure"
    with pytest.raises(ValueError):
        Flexure(grid, method="bad-name")


def test_eet_attribute():
    grid = RasterModelGrid((20, 20), xy_spacing=10e3)
    for val in (10e3, 1e3):
        assert Flexure(grid, eet=val).eet == pytest.approx(val)
    with pytest.raises(ValueError):
        assert Flexure(grid, eet=-10e3)


def test_youngs_attribute():
    grid = RasterModelGrid((20, 20), xy_spacing=10e3)
    for val in (10e3, 1e3):
        assert Flexure(grid, youngs=val).youngs == pytest.approx(val)


def test_gravity_attribute():
    grid = RasterModelGrid((20, 20), xy_spacing=10e3)
    for val in (10e3, 1e3):
        assert Flexure(grid, gravity=val).gravity == pytest.approx(val)


def test_rho_mantle_attribute():
    grid = RasterModelGrid((20, 20), xy_spacing=10e3)
    for val in (10e3, 1e3):
        assert Flexure(grid, rho_mantle=val).rho_mantle == pytest.approx(val)


def test_name(flex):
    assert flex.name == "Flexure"


def test_input_var_names(flex):
    assert flex.input_var_names == ("lithosphere__overlying_pressure_increment",)


def test_output_var_names(flex):
    assert flex.output_var_names == ("lithosphere_surface__elevation_increment",)


def test_var_units(flex):
    assert set(flex.input_var_names) | set(flex.output_var_names) == set(
        dict(flex.units).keys()
    )

    assert flex.var_units("lithosphere_surface__elevation_increment") == "m"
    assert flex.var_units("lithosphere__overlying_pressure_increment") == "Pa"


def test_grid_shape(flex):
    assert flex.grid.number_of_node_rows == _SHAPE[0]
    assert flex.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(flex):
    assert flex.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(flex):
    assert flex.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(flex):
    for name in flex.grid["node"]:
        field = flex.grid["node"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            flex.grid.number_of_node_rows * flex.grid.number_of_node_columns,
        )

    with pytest.raises(KeyError):
        flex.grid["not_a_var_name"]


def test_field_initialized_to_zero(flex):
    for name in flex.grid["node"]:
        field = flex.grid["node"][name]
        assert np.all(field == 0.)


def test_update():
    n = 11
    n_mid = (n - 1) // 2
    i_mid = np.ravel_multi_index((n_mid, n_mid), (n, n))
    load_0 = 1e9

    grid = RasterModelGrid((n, n), xy_spacing=1e3)
    flex = Flexure(grid, method="flexure")

    load = grid.at_node["lithosphere__overlying_pressure_increment"]
    load[i_mid] = load_0

    flex.update()
    dz = flex.grid.at_node["lithosphere_surface__elevation_increment"].reshape((n, n))

    assert np.argmax(dz) == i_mid
    assert dz[n_mid, n_mid] > 0.
    assert np.all(dz[:, n_mid::-1] == dz[:, n_mid:])
    assert np.all(dz[n_mid::-1, :] == dz[n_mid:, :])


def test_subside_loads():
    n, load_0 = 11, 1e9

    grid = RasterModelGrid((n, n), xy_spacing=1e3)
    flex = Flexure(grid, method="flexure")

    grid.at_node["lithosphere__overlying_pressure_increment"][0] = load_0
    flex.update()
    dz_expected = flex.grid.at_node["lithosphere_surface__elevation_increment"]

    load = np.zeros((n, n))
    load[0, 0] = load_0

    dz = flex.subside_loads(load)
    assert np.all(dz.flatten() == pytest.approx(dz_expected))

    out = np.zeros((n, n))
    dz = flex.subside_loads(load, out=out)
    assert dz is out



import pytest

from landlab import RasterModelGrid
from landlab.components import (
    FastscapeEroder,
    FlowAccumulator,
    SedDepEroder,
    StreamPowerEroder,
    StreamPowerSmoothThresholdEroder,
)


def test_route_to_multiple_error_raised_init_FastscapeEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        FastscapeEroder(mg)


def test_route_to_multiple_error_raised_init_SedDepEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        SedDepEroder(mg)


def test_route_to_multiple_error_raised_init_StreamPowerSmoothThresholdEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        StreamPowerSmoothThresholdEroder(mg)


def test_route_to_multiple_error_raised_init_StreamPowerEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        StreamPowerEroder(mg)


def test_route_to_multiple_error_raised_run_FastscapeEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    sp = FastscapeEroder(mg, K_sp=0.1)

    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        sp.run_one_step(10)


def test_route_to_multiple_error_raised_run_SedDepEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    sp = SedDepEroder(mg)

    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        sp.run_one_step(10)


def test_route_to_multiple_error_raised_run_StreamPowerSmoothThresholdEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    sp = StreamPowerSmoothThresholdEroder(mg, K_sp=0.1, use_Q=mg.ones(at="node"))

    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        sp.run_one_step(10)


def test_route_to_multiple_error_raised_run_StreamPowerEroder():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    sp = StreamPowerEroder(mg)

    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        sp.run_one_step(10)



"""Test simple stream power functionality when a discharge array is specified."""
import os

import numpy as np
from numpy.testing import assert_array_almost_equal
from six.moves import range

from landlab import ModelParameterDictionary, RasterModelGrid
from landlab.components import FlowAccumulator, StreamPowerEroder

_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_sp_discharges_old():
    input_str = os.path.join(_THIS_DIR, "test_sp_params_discharge.txt")
    inputs = ModelParameterDictionary(input_str)
    nrows = 5
    ncols = 5
    dx = inputs.read_float("dx")
    dt = inputs.read_float("dt")

    mg = RasterModelGrid(nrows, ncols, xy_spacing=dx)
    mg.add_zeros("topographic__elevation", at="node")
    z = np.array(
        [
            5.,
            5.,
            0.,
            5.,
            5.,
            5.,
            2.,
            1.,
            2.,
            5.,
            5.,
            3.,
            2.,
            3.,
            5.,
            5.,
            4.,
            4.,
            4.,
            5.,
            5.,
            5.,
            5.,
            5.,
            5.,
        ]
    )
    mg["node"]["topographic__elevation"] = z

    fr = FlowAccumulator(mg, flow_director="D8")
    my_Q = mg.at_node["surface_water__discharge"]
    sp = StreamPowerEroder(mg, input_str, use_Q=my_Q)

    # perform the loop (once!)
    for i in range(1):
        fr.run_one_step()
        my_Q[:] = mg.at_node["surface_water__discharge"] * 1.
        sp.run_one_step(dt)

    z_tg = np.array(
        [
            5.,
            5.,
            0.,
            5.,
            5.,
            5.,
            1.47759225,
            0.43050087,
            1.47759225,
            5.,
            5.,
            2.32883687,
            1.21525044,
            2.32883687,
            5.,
            5.,
            3.27261262,
            3.07175015,
            3.27261262,
            5.,
            5.,
            5.,
            5.,
            5.,
            5.,
        ]
    )

    assert_array_almost_equal(mg.at_node["topographic__elevation"], z_tg)


def test_sp_discharges_new():
    input_str = os.path.join(_THIS_DIR, "test_sp_params_discharge_new.txt")
    inputs = ModelParameterDictionary(input_str, auto_type=True)
    nrows = 5
    ncols = 5
    dx = inputs.read_float("dx")
    dt = inputs.read_float("dt")

    mg = RasterModelGrid(nrows, ncols, xy_spacing=dx)
    mg.add_zeros("topographic__elevation", at="node")
    z = np.array(
        [
            5.,
            5.,
            0.,
            5.,
            5.,
            5.,
            2.,
            1.,
            2.,
            5.,
            5.,
            3.,
            2.,
            3.,
            5.,
            5.,
            4.,
            4.,
            4.,
            5.,
            5.,
            5.,
            5.,
            5.,
            5.,
        ]
    )
    mg["node"]["topographic__elevation"] = z

    fr = FlowAccumulator(mg, flow_director="D8")
    sp = StreamPowerEroder(mg, **inputs)

    # perform the loop (once!)
    for i in range(1):
        fr.run_one_step()
        sp.run_one_step(dt)

    z_tg = np.array(
        [
            5.,
            5.,
            0.,
            5.,
            5.,
            5.,
            1.47759225,
            0.43050087,
            1.47759225,
            5.,
            5.,
            2.32883687,
            1.21525044,
            2.32883687,
            5.,
            5.,
            3.27261262,
            3.07175015,
            3.27261262,
            5.,
            5.,
            5.,
            5.,
            5.,
            5.,
        ]
    )

    assert_array_almost_equal(mg.at_node["topographic__elevation"], z_tg)




import numpy as np
from scipy.stats import fisk, genextreme
from six.moves import range

from landlab import CLOSED_BOUNDARY, Component, RasterModelGrid


class SpatialPrecipitationDistribution(Component):
    """Generate spatially resolved precipitation events.

    A component to generate a sequence of spatially resolved storms over a
    grid, following a lightly modified version (see below) of the
    stochastic methods of Singer & Michaelides, Env Res Lett 12, 104011,
    2017, & Singer et al., Geosci. Model Dev., accepted, 10.5194/gmd-2018-86.

    The method is heavily stochastic, and at the present time is intimately
    calibrated against the conditions at Walnut Gulch, described in those
    papers. In particular, assumptions around intensity-duration
    calibration and orographic rainfall are "burned in" for now, and are
    not accessible to the user. The various probability distributions
    supplied to the various run methods default to WG values, but are
    easily modified.  This calibration reflects a US desert southwest
    "monsoonal" climate, and the component distinguishes (optionally)
    between two seasons, "monsoonal" and "winter". The intensity-duration
    relationship is shared between the seasons, and so may prove useful in
    a variety of storm-dominated contexts.

    The default is to disable the orographic rainfall functionality of the
    component. However, if orographic_scenario == 'Singer', the component
    requires a 'topographic__elevation' field to already exist on the grid
    at the time of instantiation.

    The component has two ways of simulating a "year". This choice is
    controlled by the 'limit' parameter of the yield methods. If limit==
    'total_rainfall', the component will continue to run until the total
    rainfall for the season and/or year exceeds a stochastically generated
    value. This method is directly comparable to the Singer & Michaelides
    method, but will almost always result in years which are not one
    calendar year long, unless the input distributions are very carefully
    recalibrated for each use case. If limit=='total_time', the component
    will terminate a season and/or year once the elapsed time exceeds one
    year. In this case, the total rainfall will not correspond to the
    stochastically generated total. You can access the actual total for the
    last season using the property `(median_)total_rainfall_last_season`.

    Note that this component cannot simulate the occurrence of more than one
    storm at the same time. Storms that should be synchronous will instead
    occur sequentially, with no interstorm time. This limitation means that
    if enough storms occur in a year that numstorms*mean_storm_duration
    exceeds one year, the number of simulated storms will saturate. This
    limitation may be relaxed in the future.

    The component offers the option to modify the maximum number of storms
    simulated per year. If you find simulations encountering this limit too
    often, you may need to raise this limit. Conversely, it could be lowered
    to reduce memory usage over small grids. However, in increasing the value,
    beware - the component maintains two limit*nnodes arrays, which will chew
    through memory if the limit gets too high. The default will happily
    simulate grids up to around 50 km * 50 km using the default probability
    distributions.

    Key methods are:
    yield_storms
        Generate a timeseries of storm:interstorm duration pairs, alongside
        a field that describes the spatial distribution of rain during that
        storm.
    yield_years
        Generate a timeseries of ints giving number of storms per year,
        alongside a field that describes the spatial distribution of total
        rainfall across that year.
    yield_seasons
        Generate a timeseries of ints giving number of storms per season,
        alongside a field that describes the spatial distribution of total
        rainfall across that season.
    calc_annual_rainfall
        Produce a timeseries of tuples giving total rainfall each season,
        without resolving the storms spatially (i.e., fast!).

    A large number of properties are available to access storm properties
    during generation:
    - current_year
    - current_season
    - storm_depth_last_storm
    - storm_recession_value_last_storm
    - storm_duration_last_storm
    - storm_area_last_storm
    - storm_intensity_last_storm
    - total_rainfall_this_season
    - total_rainfall_this_year
    - total_rainfall_last_season
    - total_rainfall_last_year
    - median_total_rainfall_this_season
    - median_total_rainfall_this_year
    - median_total_rainfall_last_season
    - median_total_rainfall_last_year
    - number_of_nodes_under_storm
    - nodes_under_storm
    - target_median_total_rainfall_this_season

    Note that becuase these are medians not means,
    median_total_rainfall_last_season + median_total_rainfall_this_season
    != median_total_rainfall_this_year.

    Significant differences between this component and the Singer code are:
    - The component does not model evapotranspiration. Use a separate
        Landlab component for this.
    - The component runs only over a LL grid; there is no such thing as a
        validation or simulation run.
    - It produces "fuzz" around intensity values using a continuous
        distribution; Singer does this with integer steps.
    - Step changes mid-run cannot be explicitly modelled. Instead, run the
        component for a fixed duration, make the change to the
        distribution input parameter, then run it again.
    - Storms can be centred at any spatial coordinate, not just over nodes.
    - Edge buffering is now dynamic; i.e., big storms have a bigger edge
        buffer than smaller storms. Storms can be centered off the grid
        edges.
    - Storms are never discarded - once a storm is drawn, it must hit the
        catchment, and positions are repeatedly selected until this can
        happen. Singer's method would discard such a storm and draw a new
        one.
    - Durations are not rescaled to ensure both total duration and total
        precip are both satisfied at the same time, as in Singer's method.
        Instead, the component either matches a year's duration, *or*
        exactly a year's worth of rain. This choice is dictated by the
        `limit` parameter in the yield methods.

    Examples
    --------

    >>> import numpy as np
    >>> from landlab import RasterModelGrid, VoronoiDelaunayGrid
    >>> mg = RasterModelGrid((10, 10), xy_spacing=1000.)
    >>> rain = SpatialPrecipitationDistribution(mg)

    Calling yield_storms will produce storm-interstorm duration (hr) pairs
    until the model runtime has elapsed.

    >>> np.random.seed(1)
    >>> total_t_each_step = [
    ...     (storm+interstorm) for (storm, interstorm) in rain.yield_storms()]
    >>> len(total_t_each_step)
    41
    >>> np.isclose(sum(total_t_each_step)/24., 365.)
    True

    The actual rainfall intensities during that interval are accessible in the
    'rainfall__flux' field (mm/hr). The storm centre does not have to be over
    the grid, but in this case, it was for the last simulated storm:

    >>> mg.at_node['rainfall__flux'].argmax()
    80

    We can also run the component for only one season (i.e., only using one
    of the pdf sets describing the storm properties):

    >>> for field in ('rainfall__flux', 'rainfall__total_depth_per_year'):
    ...     _ = mg.at_node.pop(field)  # clear out the existing fields
    >>> rain = SpatialPrecipitationDistribution(mg, number_of_years=2)
    >>> np.random.seed(5)
    >>> total_t_each_step = [
    ...     (storm+interstorm) for (storm, interstorm) in rain.yield_storms(
    ...         style='monsoonal', monsoon_fraction_of_year=0.35)]
    >>> np.isclose(sum(total_t_each_step)/24./365./2., 0.35)
    True

    Note this behaviour can be stopped by upping monsoon_fraction_of_year:

    >>> np.random.seed(5)
    >>> total_t_each_step = [
    ...     (storm+interstorm) for (storm, interstorm) in rain.yield_storms(
    ...         style='monsoonal', monsoon_fraction_of_year=1.)]
    >>> np.isclose(round(sum(total_t_each_step)/24./365./2., 2), 1.)
    True

    yield_years yields the number of storms in the last whole year.
    Use 'rainfall__total_depth_per_year' to access the rainfall map for the
    last fully elapsed year, or equivalently, the total_rainfall_last_year
    property. Note the component seamlessly handles non-raster grid types:

    >>> vdg = VoronoiDelaunayGrid(np.random.rand(100)*1000.,
    ...                           np.random.rand(100)*1000.)
    >>> np.random.seed(3)
    >>> rain = SpatialPrecipitationDistribution(vdg, number_of_years=3)
    >>> storms_each_year = []
    >>> for total_storms in rain.yield_years(style='monsoonal',
    ...                                      total_rf_trend=0.05,
    ...                                      storminess_trend=-0.02):
    ...     storms_each_year.append(total_storms)
    ...     assert(np.all(np.equal(
    ...         vdg.at_node['rainfall__total_depth_per_year'],
    ...         rain.total_rainfall_last_year)))
    >>> sum(storms_each_year)
    11

    yield_seasons yields rainfall statistics for individual seasons. Access
    these using the various provided component properties. Note that we can
    get the component to yield a total rainfall that is calibrated to the
    supplied total_rf_gaussians if we set limit to 'total__rainfall' rather
    than 'total_time' (at the cost of exactly matching the season length):

    >>> for field in ('rainfall__flux', 'rainfall__total_depth_per_year'):
    ...     _ = mg.at_node.pop(field)  # clear out the existing fields
    >>> rain = SpatialPrecipitationDistribution(mg, number_of_years=2)
    >>> np.random.seed(5)
    >>> season_list = []
    >>> theoretical_median_rf_season = []
    >>> median_rf_season = []
    >>> median_rf_last_year = []
    >>> mean_rf_season = []
    >>> mean_rf_last_year = []
    >>> for storm_number in rain.yield_seasons(limit='total_rainfall'):
    ...     season_list.append(rain.current_season)
    ...     theoretical_median_rf_season.append(
    ...         rain.target_median_total_rainfall_this_season)
    ...     median_rf_season.append(rain.median_total_rainfall_this_season)
    ...     median_rf_last_year.append(rain.median_total_rainfall_last_year)
    ...     mean_rf_season.append(rain.total_rainfall_this_season.mean())
    ...     mean_rf_last_year.append(rain.total_rainfall_last_year.mean())
    >>> season_list == ['M', 'W', 'M', 'W']
    True
    >>> [meas > sim for (meas, sim) in zip(
    ...     median_rf_season, theoretical_median_rf_season)]  # must exceed
    [True, True, True, True]
    >>> np.isclose(median_rf_last_year[0], 0.)
    True
    >>> for season in (0, 2):  # this property must be the same in both seasons
    ...     np.isclose(median_rf_last_year[season],
    ...                median_rf_last_year[season + 1])
    True
    True

    Note that because we work here with medians, the seasonal medians don't sum
    to the year median, but the means do:

    >>> np.isclose(median_rf_last_year[2],
    ...            median_rf_season[0] + median_rf_season[1])
    False
    >>> np.isclose(mean_rf_last_year[2],
    ...            mean_rf_season[0] + mean_rf_season[1])
    True
    """

    _name = "SpatialPrecipitationDistribution"

    _cite_as = """@Article{gmd-2018-86,
        title={STORM: A simple, flexible, and parsimonious stochastic rainfall
               generator for simulating climate and climate change},
        author={Singer, M. B. and Michaelides, K. and Hobley, D. E. J.},
        journal={Geoscientific Model Development Discussions},
        volume={2018},
        pages={1--25},
        year={2018},
        url={https://www.geosci-model-dev-discuss.net/gmd-2018-86/},
        doi={10.5194/gmd-2018-86}
        }"""

    _input_var_names = ("topographic__elevation",)

    _output_var_names = ("rainfall__flux", "rainfall__total_depth_per_year")

    _var_units = {
        "topographic__elevation": "m",
        "rainfall__flux": "mm/hr",
        "rainfall__total_depth_per_year": "mm/yr",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "rainfall__flux": "node",
        "rainfall__total_depth_per_year": "node",
    }

    _var_doc = {
        "topographic__elevation": "Land surface topographic elevation",
        "rainfall__flux": "Depth of water delivered per unit time in each storm",
        "rainfall__total_depth_per_year": "Depth of water delivered in total in each model year",
    }

    def __init__(
        self, grid, number_of_years=1, orographic_scenario=None, max_numstorms=5000
    ):
        """Create the SpatialPrecipitationDistribution generator component.

        Parameters
        ----------
        grid : ModelGrid
            A Landlab model grid of any type.
        number_of_years : int
            The number of years over which to generate storms.
        orographic_scenario : {None, 'Singer', func}
            Whether to use no orographic rule, or to adopt S&M's 2017
            calibration for Walnut Gulch. Alternatively, provide a function
            here that turns the provided elevation of the storm center into
            a length-11 curve weighting to select which orographic scenario
            to apply.

        """
        self._grid = grid
        gaugecount = (grid.status_at_node != CLOSED_BOUNDARY).sum()
        self._gauge_dist_km = np.zeros(gaugecount, dtype="float")
        self._temp_dataslots1 = np.zeros(gaugecount, dtype="float")
        self._temp_dataslots2 = np.zeros(gaugecount, dtype="float")
        self._numyrs = number_of_years

        self._max_numstorms = max_numstorms
        # This is for initializing matrices. Trailing zeros are deleted from
        # matrixes at the end of the code.

        assert orographic_scenario in (None, "Singer")
        self._orographic_scenario = orographic_scenario

        # build LL fields:
        self.initialize_output_fields()
        # bind the field to the internal variable:
        self._rain_int_gauge = self.grid.at_node["rainfall__flux"]
        self._total_rf_year = self.grid.at_node["rainfall__total_depth_per_year"]

        # store some info on the open node grid extent:
        open_nodes = self.grid.status_at_node != CLOSED_BOUNDARY
        self._minx = self.grid.node_x[open_nodes].min()
        self._maxx = self.grid.node_x[open_nodes].max()
        self._miny = self.grid.node_y[open_nodes].min()
        self._maxy = self.grid.node_y[open_nodes].max()
        self._widthx = self._maxx - self._minx
        self._widthy = self._maxy - self._miny
        self._running_total_rainfall_this_year = self.grid.zeros("node")
        self._running_total_rainfall_this_season = self.grid.zeros("node")

        self._open_area = self.grid.cell_area_at_node[open_nodes].sum()
        self._scaling_to_WG = self._open_area / 275710702.
        # ^ this is the relative size of the catchment compared to WG

    def yield_storms(
        self,
        limit="total_time",
        style="whole_year",
        total_rf_trend=0.,
        storminess_trend=0.,
        monsoon_fraction_of_year=0.42,
        monsoon_total_rf_gaussian={"sigma": 64., "mu": 207.},
        monsoon_storm_duration_GEV={
            "shape": -0.570252,
            "sigma": 35.7389,
            "mu": 34.1409,
            "trunc_interval": (0., 1040.),
        },
        monsoon_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        monsoon_storm_interarrival_GEV={
            "shape": -0.807971,
            "sigma": 9.4957,
            "mu": 10.6108,
            "trunc_interval": (0., 720.),
        },
        monsoon_storm_radial_weakening_gaussian={"sigma": 0.08, "mu": 0.25},
        winter_total_rf_gaussian={"sigma": 52., "mu": 1.65},
        winter_storm_duration_fisk={
            "c": 1.0821,
            "scale": 68.4703,
            "trunc_interval": (0., 5000.),
        },
        winter_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        winter_storm_interarrival_GEV={
            "shape": 1.1131,
            "sigma": 53.2671,
            "mu": 47.4944,
            "trunc_interval": (0., 720.),
        },
        winter_storm_radial_weakening_gaussian={
            "sigma": 0.08,
            "mu": 0.25,
            "trunc_interval": (0.15, 0.67),
        },
    ):
        """
        Yield a timeseries giving the number if storms occurring each year in
        a rainfall simulation.

        All default distributions specified as parameters reflect values for
        Walnut Gulch, see Singer & Michaelides, 2017 & Singer et al, submitted.

        Parameters
        ----------
        limit : ('total_time', 'total_rainfall')
            Controls whether a season is defined based on its total rainfall
            (and can be any length), or by its duration (and can have any
            amount of rainfall). If 'total_time', monsoon_fraction_of_year
            sets the fraction of a year occupied by the monsoon.
        style : ('whole_year', 'monsoonal', 'winter')
            Controls whether the component seeks to simulate a western US-
            style "monsoonal" climate, a western US-style winter climate,
            or a full year combining both. These distributions are by default
            based on Singer et al.'s calibrations. Note if 'monsoonal',
            the total duration of a "year" will appear to be only
            `monsoon_fraction_of_year`, and the opposite for `winter`.
        total_rf_trend : float
            Controls if a drift is applied to the total rainfall distribution
            through time. If 0., no trend. If positive, rainfall totals
            increase gradually through time. If negative, they fall through
            time. S&M recommend +/- 0.07 for a realistic climate change driven
            drift at Walnut Gulch.
        storminess_trend : float
            Controls if a drift is applied to the expected intensity of
            individual storms through time. If 0., no trend. If positive,
            storms get more intense through time, if negative, less so. S&M
            recommend +/- 0.01 for a realistic climate change driven drift at
            Walnut Gulch.
        monsoon_fraction_of_year : float
            If limit == 'total_time', sets the fraction of one year occupied
            by the monsoon season. If not, ignored. Singer's monsoon runs from
            May to September, inclusive, and the default reflects this.

       monsoon_total_rf_gaussian is a normal distribution controlling the total
            rainfall expected in each year. S&M use 'mu' in {143., 271.} for
            step changes up/down in rainfall totals.
        monsoon_storm_duration_GEV is a generalised extreme value distribution
            controlling the duration of each storm. In MIN.
        monsoon_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        monsoon_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. Note that this calibration is specifically to Walnut Gulch,
            which has an area of 275 km**2. The generator directly scales this
            resulting distribution to the area ratio of Walnut Gulch to the
            open cells of the grid. This crudely accounts for the fact that
            bigger catchments will have more storms, but note that the heavy
            tail on this distribution means the default distribution shape
            will not be trustworthy for catchments with big differences in
            size from Walnut Gulch.
        monsoon_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.

        winter_total_rf_gaussian is a normal distribution controlling the total
            rainfall expected in each year. S&M use 'mu' in {143., 271.} for
            step changes up/down in rainfall totals.
        winter_storm_duration_fisk is a Fisk (i.e., log-logistic) distribution
            controlling the duration of each storm. Note this differs from the
            summer scaling. In MIN.
        winter_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        winter_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. The same considerations apply here as for the monsoonal
            interstorm equivalent.
        winter_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.

        Yields
        ------
        (storm_t, interval_t) : (float, float)
            Tuple pair of duration of a single storm, then the interstorm
            interval that follows it. In hrs. The rainfall__flux field
            describes the rainfall rate during the interval storm_t as the
            tuple is yielded. In HRS.
            Note that the rainfall__total_depth_per_year field gives the total
            accumulated rainfall depth during the *last completed* model year,
            not the year to the point of yield. For the latter, use the
            property `total_rainfall_this_year`.
        """
        return self._run_the_process(
            yield_storms=True,
            yield_years=False,
            yield_seasons=False,
            limit=limit,
            style=style,
            monsoon_fraction_of_year=monsoon_fraction_of_year,
            total_rf_trend=total_rf_trend,
            storminess_trend=storminess_trend,
            monsoon_total_rf_gaussian=monsoon_total_rf_gaussian,
            monsoon_storm_duration_GEV=monsoon_storm_duration_GEV,
            monsoon_storm_area_GEV=monsoon_storm_area_GEV,
            monsoon_storm_interarrival_GEV=monsoon_storm_interarrival_GEV,
            monsoon_storm_radial_weakening_gaussian=monsoon_storm_radial_weakening_gaussian,
            winter_total_rf_gaussian=winter_total_rf_gaussian,
            winter_storm_duration_fisk=winter_storm_duration_fisk,
            winter_storm_area_GEV=winter_storm_area_GEV,
            winter_storm_interarrival_GEV=winter_storm_interarrival_GEV,
            winter_storm_radial_weakening_gaussian=winter_storm_radial_weakening_gaussian,
        )

    def yield_years(
        self,
        limit="total_time",
        style="whole_year",
        total_rf_trend=0.,
        storminess_trend=0.,
        monsoon_fraction_of_year=0.42,
        monsoon_total_rf_gaussian={"sigma": 64., "mu": 207.},
        monsoon_storm_duration_GEV={
            "shape": -0.570252,
            "sigma": 35.7389,
            "mu": 34.1409,
            "trunc_interval": (1., 1040.),
        },
        monsoon_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        monsoon_storm_interarrival_GEV={
            "shape": -0.807971,
            "sigma": 9.4957,
            "mu": 10.6108,
            "trunc_interval": (0., 720.),
        },
        monsoon_storm_radial_weakening_gaussian={
            "sigma": 0.08,
            "mu": 0.25,
            "trunc_interval": (0.15, 0.67),
        },
        winter_total_rf_gaussian={"sigma": 52., "mu": 1.65},
        winter_storm_duration_fisk={
            "c": 1.0821,
            "scale": 68.4703,
            "trunc_interval": (1., 5000.),
        },
        winter_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        winter_storm_interarrival_GEV={
            "shape": 1.1131,
            "sigma": 53.2671,
            "mu": 47.4944,
            "trunc_interval": (0., 720.),
        },
        winter_storm_radial_weakening_gaussian={
            "sigma": 0.08,
            "mu": 0.25,
            "trunc_interval": (0.15, 0.67),
        },
    ):
        """
        Yield a timeseries giving the number if storms occurring each year in
        a rainfall simulation.

        All default distributions specified as parameters reflect values for
        Walnut Gulch, see Singer & Michaelides, 2017 & Singer et al, submitted.

        Parameters
        ----------
        limit : ('total_time', 'total_rainfall')
            Controls whether a season is defined based on its total rainfall
            (and can be any length), or by its duration (and can have any
            amount of rainfall). If 'total_time', monsoon_fraction_of_year
            sets the fraction of a year occupied by the monsoon.
        style : ('whole_year', 'monsoonal', 'winter')
            Controls whether the component seeks to simulate a western US-
            style "monsoonal" climate, a western US-style winter climate,
            or a full year combining both. These distributions are by default
            based on Singer et al.'s calibrations. Note if 'monsoonal',
            the total duration of a "year" will appear to be only
            `monsoon_fraction_of_year`, and the opposite for `winter`.
        total_rf_trend : float
            Controls if a drift is applied to the total rainfall distribution
            through time. If 0., no trend. If positive, rainfall totals
            increase gradually through time. If negative, they fall through
            time. S&M recommend +/- 0.07 for a realistic climate chage driven
            drift at Walnut Gulch.
        storminess_trend : float
            Controls if a drift is applied to the expected intensity of
            individual storms through time. If 0., no trend. If positive,
            storms get more intense through time, if negative, less so. S&M
            recommend +/- 0.01 for a realistic climate change driven drift at
            Walnut Gulch.
        monsoon_fraction_of_year : float
            If limit == 'total_time', sets the fraction of one year occupied
            by the monsoon season. If not, ignored. Singer's monsoon runs from
            May to September, inclusive, and the default reflects this.

        monsoon_total_rf_gaussian is a normal distribution controlling the
            total rainfall expected in each year. S&M use 'mu' in {143., 271.}
            for step changes up/down in rainfall totals. In mm.
        monsoon_storm_duration_GEV is a generalised extreme value distribution
            controlling the duration of each storm. In MIN.
        monsoon_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        monsoon_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. Note that this calibration is specifically to Walnut Gulch,
            which has an area of 275 km**2. The generator directly scales this
            resulting distribution to the area ratio of Walnut Gulch to the
            open cells of the grid. This crudely accounts for the fact that
            bigger catchments will have more storms, but note that the heavy
            tail on this distribution means the default distribution shape
            will not be trustworthy for catchments with big differences in
            size from Walnut Gulch.
        monsoon_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.

        winter_total_rf_gaussian is a normal distribution controlling the total
            rainfall expected in each year. S&M use 'mu' in {143., 271.} for
            step changes up/down in rainfall totals.
        winter_storm_duration_fisk is a Fisk (i.e., log-logistic) distribution
            controlling the duration of each storm. Note this differs from the
            summer scaling. In MIN.
        winter_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        winter_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. The same considerations apply here as for the monsoonal
            interstorm equivalent.
        winter_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.

        Yields
        ------
        number_of_storms_per_year : float
            Float that gives the number of storms simulated in the year that
            elapsed since the last yield. The rainfall__total_depth_per_year
            field gives the total accumulated rainfall depth during the year
            preceding the yield. rainfall__flux gives the rainfall intensity of
            the last storm in that year.
        """
        return self._run_the_process(
            yield_storms=False,
            yield_years=True,
            yield_seasons=False,
            limit=limit,
            style=style,
            total_rf_trend=total_rf_trend,
            storminess_trend=storminess_trend,
            monsoon_fraction_of_year=monsoon_fraction_of_year,
            monsoon_total_rf_gaussian=monsoon_total_rf_gaussian,
            monsoon_storm_duration_GEV=monsoon_storm_duration_GEV,
            monsoon_storm_area_GEV=monsoon_storm_area_GEV,
            monsoon_storm_interarrival_GEV=monsoon_storm_interarrival_GEV,
            monsoon_storm_radial_weakening_gaussian=monsoon_storm_radial_weakening_gaussian,
            winter_total_rf_gaussian=winter_total_rf_gaussian,
            winter_storm_duration_fisk=winter_storm_duration_fisk,
            winter_storm_area_GEV=winter_storm_area_GEV,
            winter_storm_interarrival_GEV=winter_storm_interarrival_GEV,
            winter_storm_radial_weakening_gaussian=winter_storm_radial_weakening_gaussian,
        )

    def yield_seasons(
        self,
        limit="total_time",
        style="whole_year",
        total_rf_trend=0.,
        storminess_trend=0.,
        monsoon_fraction_of_year=0.42,
        monsoon_total_rf_gaussian={"sigma": 64., "mu": 207.},
        monsoon_storm_duration_GEV={
            "shape": -0.570252,
            "sigma": 35.7389,
            "mu": 34.1409,
            "trunc_interval": (1., 1040.),
        },
        monsoon_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        monsoon_storm_interarrival_GEV={
            "shape": -0.807971,
            "sigma": 9.4957,
            "mu": 10.6108,
            "trunc_interval": (0., 720.),
        },
        monsoon_storm_radial_weakening_gaussian={
            "sigma": 0.08,
            "mu": 0.25,
            "trunc_interval": (0.15, 0.67),
        },
        winter_total_rf_gaussian={"sigma": 52., "mu": 1.65},
        winter_storm_duration_fisk={
            "c": 1.0821,
            "scale": 68.4703,
            "trunc_interval": (1., 5000.),
        },
        winter_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        winter_storm_interarrival_GEV={
            "shape": 1.1131,
            "sigma": 53.2671,
            "mu": 47.4944,
            "trunc_interval": (0., 720.),
        },
        winter_storm_radial_weakening_gaussian={
            "sigma": 0.08,
            "mu": 0.25,
            "trunc_interval": (0.15, 0.67),
        },
    ):
        """
        Yield a timeseries giving the number if storms occurring each season in
        a rainfall simulation. Only meaningfully different from yield_years if
        style=='whole_year'.

        All default distributions specified as parameters reflect values for
        Walnut Gulch, see Singer & Michaelides, 2017 & Singer et al, submitted.

        Parameters
        ----------
        limit : ('total_time', 'total_rainfall')
            Controls whether a season is defined based on its total rainfall
            (and can be any length), or by its duration (and can have any
            amount of rainfall). If 'total_time', monsoon_fraction_of_year
            sets the fraction of a year occupied by the monsoon.
        style : ('whole_year', 'monsoonal', 'winter')
            Controls whether the component seeks to simulate a western US-
            style "monsoonal" climate, a western US-style winter climate,
            or a full year combining both. These distributions are by default
            based on Singer et al.'s calibrations. Note if 'monsoonal',
            the total duration of a "year" will appear to be only
            `monsoon_fraction_of_year`, and the opposite for `winter`.
        total_rf_trend : float
            Controls if a drift is applied to the total rainfall distribution
            through time. If 0., no trend. If positive, rainfall totals
            increase gradually through time. If negative, they fall through
            time. S&M recommend +/- 0.07 for a realistic climate chage driven
            drift at Walnut Gulch.
        storminess_trend : float
            Controls if a drift is applied to the expected intensity of
            individual storms through time. If 0., no trend. If positive,
            storms get more intense through time, if negative, less so. S&M
            recommend +/- 0.01 for a realistic climate change driven drift at
            Walnut Gulch.
        monsoon_fraction_of_year : float
            If limit == 'total_time', sets the fraction of one year occupied
            by the monsoon season. If not, ignored. Singer's monsoon runs from
            May to September, inclusive, and the default reflects this.

        monsoon_total_rf_gaussian is a normal distribution controlling the
            total rainfall expected in each year. S&M use 'mu' in {143., 271.}
            for step changes up/down in rainfall totals. In mm.
        monsoon_storm_duration_GEV is a generalised extreme value distribution
            controlling the duration of each storm. In MIN.
        monsoon_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        monsoon_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. Note that this calibration is specifically to Walnut Gulch,
            which has an area of 275 km**2. The generator directly scales this
            resulting distribution to the area ratio of Walnut Gulch to the
            open cells of the grid. This crudely accounts for the fact that
            bigger catchments will have more storms, but note that the heavy
            tail on this distribution means the default distribution shape
            will not be trustworthy for catchments with big differences in
            size from Walnut Gulch.
        monsoon_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.

        winter_total_rf_gaussian is a normal distribution controlling the total
            rainfall expected in each year. S&M use 'mu' in {143., 271.} for
            step changes up/down in rainfall totals.
        winter_storm_duration_fisk is a Fisk (i.e., log-logistic) distribution
            controlling the duration of each storm. Note this differs from the
            summer scaling. In MIN.
        winter_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        winter_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. The same considerations apply here as for the monsoonal
            interstorm equivalent.
        winter_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.

        Yields
        ------
        number_of_storms_per_season : float
            Float that gives the number of storms simulated in the season that
            elapsed since the last yield. The rainfall__total_depth_per_year
            field gives the total accumulated rainfall depth during the *year*
            preceding the yield, *so far*. rainfall__flux gives the rainfall
            intensity of the last storm in that year.
        NB: Use the component property total_rainfall_last_season to access
            the *actual* amount of rainfall in the season that has the number
            of storms that the method generates.
        """
        return self._run_the_process(
            yield_storms=False,
            yield_years=False,
            yield_seasons=True,
            limit=limit,
            style=style,
            total_rf_trend=total_rf_trend,
            storminess_trend=storminess_trend,
            monsoon_fraction_of_year=monsoon_fraction_of_year,
            monsoon_total_rf_gaussian=monsoon_total_rf_gaussian,
            monsoon_storm_duration_GEV=monsoon_storm_duration_GEV,
            monsoon_storm_area_GEV=monsoon_storm_area_GEV,
            monsoon_storm_interarrival_GEV=monsoon_storm_interarrival_GEV,
            monsoon_storm_radial_weakening_gaussian=monsoon_storm_radial_weakening_gaussian,
            winter_total_rf_gaussian=winter_total_rf_gaussian,
            winter_storm_duration_fisk=winter_storm_duration_fisk,
            winter_storm_area_GEV=winter_storm_area_GEV,
            winter_storm_interarrival_GEV=winter_storm_interarrival_GEV,
            winter_storm_radial_weakening_gaussian=winter_storm_radial_weakening_gaussian,
        )

    def _run_the_process(
        self,
        yield_storms=True,
        yield_years=False,
        yield_seasons=False,
        limit="total_time",
        style="whole_year",
        monsoon_fraction_of_year=0.42,
        total_rf_trend=0.,
        storminess_trend=0.,
        monsoon_total_rf_gaussian={"sigma": 64., "mu": 207.},
        monsoon_storm_duration_GEV={
            "shape": -0.570252,
            "sigma": 35.7389,
            "mu": 34.1409,
            "trunc_interval": (1., 1040.),
        },
        monsoon_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        monsoon_storm_interarrival_GEV={
            "shape": -0.807971,
            "sigma": 9.4957,
            "mu": 10.6108,
            "trunc_interval": (0., 720.),
        },
        monsoon_storm_radial_weakening_gaussian={
            "sigma": 0.08,
            "mu": 0.25,
            "trunc_interval": (0.15, 0.67),
        },
        winter_total_rf_gaussian={"sigma": 52., "mu": 1.65},
        winter_storm_duration_fisk={
            "c": 1.0821,
            "scale": 68.4703,
            "trunc_interval": (1., 5000.),
        },
        winter_storm_area_GEV={
            "shape": 0.,
            "sigma": 2.83876e+07,
            "mu": 1.22419e+08,
            "trunc_interval": (5.e+06, 3.e+08),
        },
        winter_storm_interarrival_GEV={
            "shape": 1.1131,
            "sigma": 53.2671,
            "mu": 47.4944,
            "trunc_interval": (0., 720.),
        },
        winter_storm_radial_weakening_gaussian={
            "sigma": 0.08,
            "mu": 0.25,
            "trunc_interval": (0.15, 0.67),
        },
    ):
        """
        This is the underlying process that runs the component, but it should
        be run by a user through the yield_storms and yield_years methods.

        Fuzz to the chosen values is now selected from a continuous
        distribution, not from integer values.

        total_rf_trend controls if a drift is applied to the total rainfall
        distribution through time. If 0., no trend. If positive, rainfall
        totals increase gradually through time. If negative, they fall through
        time. S&M recommend +/- 0.07 for a realistic climate chage driven drift
        at Walnut Gulch.

        storminess_trend controls if a drift is applied to the expected
        intensity of individual storms through time. If 0., no trend. If
        positive, storms get more intense through time, if negative, less so.
        S&M recommend +/- 0.01 for a realistic climate change driven drift at
        Walnut Gulch.

        All default distributions reflect values for Walnut Gulch, see Singer &
        Michaelides, submitted:

        monsoon_total_rf_gaussian is a normal distribution controlling the
            total rainfall expected in each year. S&M use 'mu' in {143., 271.}
            for step changes up/down in rainfall totals. In mm.
        monsoon_storm_duration_GEV is a generalised extreme value distribution
            controlling the duration of each storm. In MIN.
        monsoon_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        monsoon_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. Note that this calibration is specifically to Walnut Gulch,
            which has an area of 275 km**2. The generator directly scales this
            resulting distribution to the area ratio of Walnut Gulch to the
            open cells of the grid. This crudely accounts for the fact that
            bigger catchments will have more storms, but note that the heavy
            tail on this distribution means the default distribution shape
            will not be trustworthy for catchments with big differences in
            size from Walnut Gulch.
        monsoon_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.
        winter_total_rf_gaussian is a normal distribution controlling the total
            rainfall expected in each year. S&M use 'mu' in {143., 271.} for
            step changes up/down in rainfall totals.
        winter_storm_duration_fisk is a Fisk (i.e., log-logistic) distribution
            controlling the duration of each storm. Note this differs from the
            summer scaling. In MIN.
        winter_storm_area_GEV is a generalised extreme value distribution
            controlling the plan view area of each storm. S&M use 'shape': 0.,
            which collapses the distribution to a plain extreme value
            distribution.
        winter_storm_interarrival_GEV is a generalised extreme value
            distribution controlling the interarrival time between each storm.
            In HRS. The same considerations apply here as for the monsoonal
            interstorm equivalent.
        winter_storm_radial_weakening_gaussian is a normal distribution
            controlling the rate of intensity decline with distance from storm
            center. For more detail see Rodriguez-Iturbe et al., 1986; Morin
            et al., 2005.

        """
        FUZZMETHOD = "DEJH"
        FUZZWIDTH = 5.  # if DEJH
        self._phantom_storm_count = 0
        # ^this property tracks the number of storms in the run that received
        # zero intensity (and thus didn't really exist)
        self._opennodes = self.grid.status_at_node != CLOSED_BOUNDARY
        self._total_rainfall_last_season = self.grid.zeros("node")

        # safety check for init conds:
        if yield_storms:
            assert yield_years is False
            assert yield_seasons is False
        if yield_years:
            assert yield_storms is False
            assert yield_seasons is False
        if yield_seasons:
            assert yield_storms is False
            assert yield_years is False

        # add variable for number of simulations of simyears
        simyears = self._numyrs  # number of years to simulate
        numcurves = 11  # number of intensity-duration curves (see below for
        # curve equations)
        hrsinyr = 24. * 365.
        hrsinmonsoon = monsoon_fraction_of_year * hrsinyr
        hrsinwinter = (1. - monsoon_fraction_of_year) * hrsinyr

        assert limit in ("total_rainfall", "total_time")

        assert style in ("whole_year", "monsoonal", "winter")
        if style == "whole_year":
            reps = 2
        else:
            reps = 1

        opennodes = self._opennodes
        num_opennodes = np.sum(opennodes)
        IDs_open = np.where(opennodes)[0]  # need this later
        X1 = self.grid.node_x
        Y1 = self.grid.node_y
        Xin = X1[opennodes]
        Yin = Y1[opennodes]
        try:
            Zz = self.grid.at_node["topographic__elevation"][opennodes]
        except KeyError:
            assert self._orographic_scenario is None
        numgauges = Xin.size  # number of rain gauges in the basin.
        # NOTE: In this version this produces output on a grid, rather than at
        # real gauge locations.

        assert FUZZMETHOD == "DEJH", "The Singer method for fuzz is no longer supported"

        # lambda_, kappa, and C are parameters of the intensity-duration curves
        # of the form: intensity =
        # lambda*exp(-0.508*duration)+kappa*exp(-0.008*duration)+C
        lambda_ = [
            642.2,
            578.0,
            513.8,
            449.5,
            385.3,
            321.1,
            256.9,
            192.7,
            128.4,
            64.1,
            21.0,
        ]
        kappa = [93.1, 83.8, 74.5, 65.2, 55.9, 46.6, 37.2, 27.9, 18.6, 9.3, 0.9]
        C = [4.5, 4., 3.5, 3., 2.5, 2., 1.5, 1., 0.5, 0.25, 0.05]

        # Unlike MS's original implementation, we no longer pull ET values, as
        # this should be a different component.

        self._Ptot_ann_global = np.zeros(simyears)
        self._Ptot_monsoon_global = np.zeros(simyears)

        master_storm_count = 0
        storm_trend = 0

        for syear in range(simyears):
            self._year = syear
            year_time = 0.  # tracks simulation time per year in hours
            storm_trend += storminess_trend
            year_storm_count = 0
            breaker = False
            Storm_total_local_year = np.zeros((self._max_numstorms, num_opennodes))
            self._storm_running_sum_of_seasons = np.zeros(num_opennodes)
            self._storm_running_sum_1st_seas = np.zeros(num_opennodes)

            storms_yr_so_far = 0
            for seas in range(reps):
                seas_time = 0.  # tracks elapsed season time in hours
                Storm_running_sum_seas = np.zeros((2, num_opennodes))
                # ^ 1st col is running total, 2nd is data to add to it
                if seas == 0 and not style == "winter":
                    self._current_season = "M"
                    # This is the pdf fitted to all available station precip
                    # data (normal dist). It will be sampled below.
                    Ptot_pdf_norm = monsoon_total_rf_gaussian

                    # This is the pdf fitted to all available station duration
                    # data (GEV dist). It will be sampled below.
                    # #### matlab's GEV is (shape_param, scale(sigma), pos(mu))
                    # note that in Scipy, we must add a minus to the shape
                    # param for a GEV to match Matlab's implementation
                    Duration_pdf = monsoon_storm_duration_GEV
                    # This is the pdf fitted to all available station area
                    # data (EV dist). It will be sampled below.
                    # #### matlab's EV is (mu, sigma)
                    Area_pdf_EV = monsoon_storm_area_GEV
                    # This is the pdf fitted to all available station area
                    # data (GEV dist). It will be sampled below.
                    Int_arr_pdf_GEV = monsoon_storm_interarrival_GEV
                    # This is the pdf of storm gradient recession coefficients
                    # from Morin et al, 2005 (normal dist). It will be sampled
                    # below.
                    Recess_pdf_norm = monsoon_storm_radial_weakening_gaussian
                    seas_total = hrsinmonsoon
                else:
                    self._current_season = "W"
                    Ptot_pdf_norm = winter_total_rf_gaussian
                    Duration_pdf = winter_storm_duration_fisk
                    Area_pdf_EV = winter_storm_area_GEV
                    Int_arr_pdf_GEV = winter_storm_interarrival_GEV
                    Recess_pdf_norm = winter_storm_radial_weakening_gaussian
                    seas_total = hrsinwinter

                if not np.isclose(total_rf_trend, 0.):
                    mu = Ptot_pdf_norm.pop("mu")
                    mu += mu * total_rf_trend
                    Ptot_pdf_norm["mu"] = mu
                # sample from normal distribution and saves global value of
                # Ptot (that must be equalled or exceeded) for each year
                season_rf_limit = self.calc_annual_rainfall(
                    style=style, monsoon_total_rf_gaussian=Ptot_pdf_norm
                )[seas]
                self._season_rf_limit = season_rf_limit
                self._Ptot_ann_global[syear] += season_rf_limit
                if seas == 0 and not style == "winter":
                    self._Ptot_monsoon_global[syear] = season_rf_limit
                Storm_total_local_seas = np.zeros((self._max_numstorms, num_opennodes))
                seas_cum_Ptot_gauge = np.zeros(numgauges)
                self._entries = 0
                seas_storm_count = 0

                for storm in range(self._max_numstorms):
                    self._rain_int_gauge.fill(0.)
                    int_arr_val = genextreme.rvs(
                        c=Int_arr_pdf_GEV["shape"],
                        loc=Int_arr_pdf_GEV["mu"],
                        scale=Int_arr_pdf_GEV["sigma"],
                    )
                    try:
                        int_arr_val = np.clip(
                            int_arr_val,
                            Int_arr_pdf_GEV["trunc_interval"][0],
                            Int_arr_pdf_GEV["trunc_interval"][1],
                        )
                    except KeyError:
                        # ...just in case
                        if int_arr_val < 0.:
                            int_arr_val = 0.
                    # now, correct the scaling relative to WG
                    int_arr_val /= self._scaling_to_WG
                    self._int_arr_val = int_arr_val
                    # ^Samples from distribution of interarrival times (hr).
                    # This can be used to develop STORM output for use in
                    # rainfall-runoff models or any water balance application.
                    # sample uniformly from storm center matrix from grid w
                    # 10 m spacings covering basin:

                    area_val = genextreme.rvs(
                        c=Area_pdf_EV["shape"],
                        loc=Area_pdf_EV["mu"],
                        scale=Area_pdf_EV["sigma"],
                    )
                    try:
                        area_val = np.clip(
                            area_val,
                            Area_pdf_EV["trunc_interval"][0],
                            Area_pdf_EV["trunc_interval"][1],
                        )
                    except KeyError:
                        # ...just in case
                        if area_val < 0.:
                            area_val = 0.
                    self._area_val = area_val
                    # ^Samples from distribution of storm areas

                    r = np.sqrt(area_val / np.pi)  # value here shd be selected
                    rsq = r ** 2
                    # based on area above in meters to match the UTM values

                    # This way of handling storm locations is really quite
                    # different to MS's. He uses a fixed buffer width, and
                    # throws away any storm that doesn't intersect. We
                    # instead retain all storms, and *make sure* the storm
                    # intersects using a dynamic buffer. MS's method will
                    # preferentially sample larger storms, though unclear
                    # what that would mean in practice.
                    # MS also snaps his storms onto the grid. This seems
                    # unnecessary, and we don't do it here.
                    while 1:
                        cx, cy = self._locate_storm(r)
                        # Determine which gauges are hit by Euclidean geometry:
                        gdist = (Xin - cx) ** 2 + (Yin - cy) ** 2
                        mask_name = gdist <= rsq  # this is defacto MS's aa
                        # this short circuits the storm loop in the case that
                        # the storm does not affect any 'gauging' location
                        if np.any(np.equal(mask_name, True)):
                            break

                    self._x = cx
                    self._y = cy
                    year_storm_count += 1
                    seas_storm_count += 1
                    master_storm_count += 1

                    # This routine below determines to which orographic group
                    # the closest gauge to the storm center belongs to, and
                    # censors the number of curves accordingly
                    # missing top curve in GR1, top and bottom curves for GR2,
                    # and bottom curve for GR3
                    # NOTE again, DEJH thinks this could be generalised a lot

                    # original curve# probs for 30%-20%-10%: [0.0636, 0.0727,
                    # 0.0819, 0.0909, 0.0909, 0.0909, 0.0909, 0.0909, 0.1001,
                    # 0.1090, 0.1182]
                    # original curve# probs are modified as below
                    # add weights to reflect reasonable probabilities that
                    # favor lower curves:
                    if self._orographic_scenario is not None:
                        # this routine below allows for orography in precip by
                        # first determining the closest gauge and then
                        # determining its orographic grouping
                        cc = np.argmin(gdist)
                        closest_gauge_z = Zz[cc]  # this will be
                        # compared against orographic gauge groupings to
                        # determine the appropriate set of intensity-duration
                        # curves
                        if self._orographic_scenario == "Singer":
                            wgts = Singer_orographic_rainfall(closest_gauge_z)
                        else:
                            wgts = self._orographic_scenario(closest_gauge_z)
                    elif self._orographic_scenario is None:
                        wgts = [
                            0.0636,
                            0.0727,
                            0.0819,
                            0.0909,
                            0.0909,
                            0.0909,
                            0.0909,
                            0.0909,
                            0.1001,
                            0.1090,
                            0.1182,
                        ]
                    if seas == 0 and not style == "winter":
                        duration_val = genextreme.rvs(
                            c=Duration_pdf["shape"],
                            loc=Duration_pdf["mu"],
                            scale=Duration_pdf["sigma"],
                        )
                    else:
                        duration_val = fisk.rvs(
                            c=Duration_pdf["c"], scale=Duration_pdf["scale"]
                        )
                    # hacky fix to prevent occasional < 0 values:
                    # (I think because Matlab is able to set limits manually)
                    try:
                        duration_val = np.clip(
                            duration_val,
                            Duration_pdf["trunc_interval"][0],
                            Duration_pdf["trunc_interval"][1],
                        )
                    except KeyError:
                        # ...just in case
                        if duration_val < 0.:
                            duration_val = 0.
                    durationhrs = duration_val / 60.
                    self._durationhrs = durationhrs
                    year_time += durationhrs
                    seas_time += durationhrs
                    # we will always do the next storm, even if it exceeds the
                    # specified "total" time

                    # which curve did we pick?:
                    int_dur_curve_val = np.random.choice(numcurves, p=wgts)

                    intensity_val = (
                        lambda_[int_dur_curve_val] * np.exp(-0.508 * duration_val)
                        + kappa[int_dur_curve_val] * np.exp(-0.008 * duration_val)
                        + C[int_dur_curve_val]
                    )
                    # ...these curves are based on empirical data from WG

                    # this dist should look identical, w/o discretisation
                    fuzz_int_val = FUZZWIDTH * 2. * (np.random.rand() - 0.5)

                    intensity_val += fuzz_int_val
                    # ^this allows for specified fuzzy tolerance around
                    # selected intensity (but it can go -ve)
                    # formerly, here MS used a rounding and threshold to
                    # prevent storms with a value < 1. We're going to remove
                    # the rounding and threshold at zero instead. (below)

                    # This scales the storm center intensity upward, so the
                    # values at each gauge are realistic once the gradient is
                    # applied.
                    intensity_val += intensity_val * storm_trend
                    # storminess trend is applied and its effect rises each
                    # year of simulation
                    # DEJH has removed the rounding
                    # Note that is is now possible for intensity_val to go
                    # negative, so:
                    if intensity_val < 0.:
                        intensity_val = 0.
                        self._phantom_storm_count += 1
                    # note storms of zero intensity are now permitted (though
                    # should hopefully remain pretty rare.)
                    self._intensity_val = intensity_val

                    # area to determine which gauges are hit:
                    recess_val = np.random.normal(
                        loc=Recess_pdf_norm["mu"], scale=Recess_pdf_norm["sigma"]
                    )
                    try:
                        recess_val = np.clip(
                            recess_val,
                            Recess_pdf_norm["trunc_interval"][0],
                            Recess_pdf_norm["trunc_interval"][1],
                        )
                    except KeyError:
                        pass  # this one is OK <0., I think
                    self._recess_val = recess_val
                    # this pdf of recession coefficients determines how
                    # intensity declines with distance from storm center (see
                    # below)
                    # determine cartesian distances to all hit gauges and
                    # associated intensity values at each gauge hit by the
                    # storm
                    # This is a data storage solution to avoid issues that can
                    # arise with slicing grid areas with heavy tailed sizes
                    self._entries = np.sum(mask_name)  # only open nodes
                    entries = self._entries
                    # NOTE _gauge_dist_km only contains nodes under the storm!
                    # The remaining entries are garbage
                    # Xin -> only the open nodes, note
                    self._gauge_dist_km[:entries] = np.sqrt(gdist[mask_name]) / 1000.
                    self._temp_dataslots2[:entries] = gdist[mask_name] / 1.e6
                    self._temp_dataslots2[:entries] *= -2. * recess_val ** 2
                    np.exp(
                        self._temp_dataslots2[:entries],
                        out=self._temp_dataslots2[:entries],
                    )
                    self._temp_dataslots2[:entries] *= intensity_val
                    mask_incl_closed = IDs_open[mask_name]
                    self._nodes_hit = mask_incl_closed
                    # ^note this is by ID, not bool
                    self._rain_int_gauge[mask_incl_closed] = self._temp_dataslots2[
                        :entries
                    ]
                    # calc of _rain_int_gauge follows Rodriguez-Iturbe et al.,
                    # 1986; Morin et al., 2005 but sampled from a distribution
                    # only need to add the bit that got rained on, so:
                    self._temp_dataslots2[:entries] *= duration_val / 60.
                    seas_cum_Ptot_gauge[mask_name] += self._temp_dataslots2[:entries]
                    # collect storm totals for all gauges into rows by storm
                    Storm_total_local_seas[storm, :] = (
                        self._rain_int_gauge[opennodes] * duration_val / 60.
                    )
                    Storm_total_local_year[
                        (storm + storms_yr_so_far), :
                    ] = Storm_total_local_seas[storm, :]
                    self._max_storm_depth = Storm_total_local_seas[storm, :].max()

                    self._Storm_total_local_seas = Storm_total_local_seas
                    self._Storm_total_local_year = Storm_total_local_year
                    Storm_running_sum_seas[1, :] = Storm_total_local_seas[storm, :]
                    np.nansum(
                        Storm_running_sum_seas, axis=0, out=Storm_running_sum_seas[0, :]
                    )
                    if np.any(Storm_total_local_seas < 0.):
                        raise ValueError(syear, storm)
                    self._median_seas_rf_total = np.nanmedian(
                        Storm_running_sum_seas[0, :]
                    )
                    self._Storm_running_sum_seas = Storm_running_sum_seas[0, :]

                    if limit == "total_time":
                        if seas_time + int_arr_val > seas_total:
                            int_arr_val = (seas_total - seas_time).clip(0.)
                            breaker = True
                    else:
                        if self._median_seas_rf_total > season_rf_limit:
                            breaker = True
                    if yield_storms is True:
                        yield (durationhrs, int_arr_val)
                    seas_time += int_arr_val
                    year_time += int_arr_val
                    if breaker:
                        # Don't create Ptotal_local per MS... just
                        breaker = False
                        break
                    if storm + 1 == self._max_numstorms:
                        raise ValueError("_max_numstorms set too low for this run")
                storms_yr_so_far = seas_storm_count
                self._storm_running_sum_of_seasons += Storm_running_sum_seas[0, :]
                self._total_rainfall_last_season[
                    self._opennodes
                ] = Storm_running_sum_seas[0, :]
                self._storm_running_sum_1st_seas += Storm_running_sum_seas[0, :]
                if yield_seasons is True:
                    yield seas_storm_count

            self._total_rf_year[opennodes] = self._storm_running_sum_of_seasons
            if yield_years is True and yield_seasons is False:
                yield year_storm_count

    def calc_annual_rainfall(
        self,
        style="whole_year",
        monsoon_total_rf_gaussian={"sigma": 64., "mu": 207.},
        winter_total_rf_gaussian={"sigma": 52., "mu": 1.65},
    ):
        """
        Return a tuple of rainfall totals (mm) for the year, with entries
        subdividing the yearly total into seasons as appropriate.

        Parameters
        ----------
        style : ('whole_year', 'monsoonal', 'winter')
            Whether to simulate 2 seasons, or a single season.
        monsoon_total_rf_gaussian : dict of sigma and mu for the summer
            distribution (if used). Defaults to Walnut Gulch.
        winter_total_rf_gaussian : dict of sigma and mu for the summer
            distribution (if used). Defaults to Walnut Gulch.

        Returns
        -------
        tuple : (first_season_total, [second_season_total])
            If style=='monsoonal' or 'winter', a len(1) tuple of the total rf.
            If style=='whole_year', a len(2) tuple of (monsoon, winter) totals.

        Examples
        --------
        >>> mg = RasterModelGrid((10, 10), xy_spacing=500.)
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> rain = SpatialPrecipitationDistribution(mg)
        >>> mytotals = []
        >>> for yr in range(5):
        ...     mytotals.append(rain.calc_annual_rainfall(style='whole_year'))
        >>> [len(x) == 2 for x in mytotals]
        [True, True, True, True, True]
        >>> mytotals = []
        >>> for yr in range(3):
        ...     mytotals.append(rain.calc_annual_rainfall(style='monsoonal'))
        >>> [len(x) == 1 for x in mytotals]
        [True, True, True]
        """
        assert style in ("whole_year", "monsoonal", "winter")
        if style in ("whole_year", "monsoonal"):
            # sample from normal distribution and saves global value of Ptot
            # (that must be equalled or exceeded) for each year
            summer_rf_limit = np.random.normal(
                loc=monsoon_total_rf_gaussian["mu"],
                scale=monsoon_total_rf_gaussian["sigma"],
            )
            try:
                summer_rf_limit = np.clip(
                    summer_rf_limit,
                    monsoon_total_rf_gaussian["trunc_interval"][0],
                    monsoon_total_rf_gaussian["trunc_interval"][1],
                )
            except KeyError:
                # ...just in case
                if summer_rf_limit < 0.:
                    summer_rf_limit = 0.
        if style in ("whole_year", "winter"):
            # sample from normal distribution and saves global value of Ptot
            # (that must be equalled or exceeded) for each year
            winter_rf_limit = np.random.normal(
                loc=winter_total_rf_gaussian["mu"],
                scale=winter_total_rf_gaussian["sigma"],
            )
            try:
                winter_rf_limit = np.clip(
                    winter_rf_limit,
                    winter_total_rf_gaussian["trunc_interval"][0],
                    winter_total_rf_gaussian["trunc_interval"][1],
                )
            except KeyError:
                # ...just in case
                if winter_rf_limit < 0.:
                    winter_rf_limit = 0.
        if style == "monsoonal":
            return (summer_rf_limit,)
        elif style == "winter":
            return (winter_rf_limit,)
        else:
            return (summer_rf_limit, winter_rf_limit)

    def _locate_storm(self, storm_radius):
        """
        Because of the way the stats fall out, any simulated storm from the
        distribution must intersect the catchment somewhere. Note written in a
        grid-agnostic fashion.
        """
        stormposx = np.random.rand() * (self._widthx + 2. * storm_radius)
        stormposy = np.random.rand() * (self._widthy + 2. * storm_radius)
        stormx = self._minx - storm_radius + stormposx
        stormy = self._miny - storm_radius + stormposy
        return stormx, stormy

    @property
    def current_year(self):
        """Get the current year as an int."""
        return self._year

    @property
    def current_season(self):
        """Get the current season. 'M' is monsoon, 'W' is winter."""
        return self._current_season

    @property
    def storm_depth_last_storm(self):
        """Get the maximum storm depth during the last storm (mm)."""
        return self._max_storm_depth

    @property
    def storm_recession_value_last_storm(self):
        """Get the recession parameter (radial die-off) for the last storm."""
        return self._recess_val

    @property
    def storm_duration_last_storm(self):
        """Get the duration (in hrs) of the last storm."""
        return self._durationhrs

    @property
    def storm_area_last_storm(self):
        """Get the area (in m**2) of the last storm."""
        return self._area_val

    @property
    def storm_intensity_last_storm(self):
        """
        Get the intensity (mm/hr) of the last storm, averaged under the storm
        footprint. Note that duration * intensity != storm max depth.
        """
        return self._intensity_val

    @property
    def total_rainfall_last_season(self):
        """
        Get the total recorded rainfall over the last (completed) simulated
        season, spatially resolved (mm).
        """
        return self._total_rainfall_last_season

    @property
    def total_rainfall_last_year(self):
        """
        Get the total recorded rainfall over the last (completed) simulated
        year, spatially resolved (mm). Equivalent to the field
        'rainfall__total_depth_per_year'.
        """
        return self._total_rf_year

    @property
    def total_rainfall_this_season(self):
        """
        Get the accumulated, spatially resolved total rainfall over the
        grid for the season so far (mm).
        """
        self._running_total_rainfall_this_season[
            self._opennodes
        ] = self._Storm_running_sum_seas
        return self._running_total_rainfall_this_season

    @property
    def total_rainfall_this_year(self):
        """
        Get the accumulated, spatially resolved total rainfall over the
        grid for the year so far (mm).
        """
        self._running_total_rainfall_this_year[self._opennodes] = (
            self._storm_running_sum_1st_seas + self._Storm_running_sum_seas
        )
        return self._running_total_rainfall_this_year

    @property
    def median_total_rainfall_last_season(self):
        """
        Get the median total rainfall recorded over the open nodes of the grid
        during the last (completed) simulated season (mm).
        """
        return np.nanmedian(self._total_rainfall_last_season[self._opennodes])

    @property
    def median_total_rainfall_last_year(self):
        """
        Get the median total rainfall recorded over the open nodes of the grid
        during the last (completed) simulated year (mm).
        """
        return np.nanmedian(self.total_rainfall_last_year[self._opennodes])

    @property
    def median_total_rainfall_this_season(self):
        """
        Get the accumulated median total rainfall over the open nodes of the
        grid so far this season (mm).
        """
        return self._median_seas_rf_total

    @property
    def median_total_rainfall_this_year(self):
        """
        Get the accumulated median total rainfall over the open nodes of the
        grid so far this year (mm).
        """
        return np.nanmedian(self.total_rainfall_this_year[self._opennodes])

    @property
    def number_of_nodes_under_storm(self):
        """Get the number of nodes under the last storm."""
        return self._entries

    @property
    def nodes_under_storm(self):
        """Get the IDs of the nodes under the last storm."""
        return self._nodes_hit

    @property
    def coordinates_of_last_storm_center(self):
        """
        Get the coordinates of the center of the last storm as (x, y).
        """
        return (self._x, self._y)

    @property
    def target_median_total_rainfall_this_season(self):
        """
        Get the stochastically generated "target" average total rainfall
        amount over the catchment for the current season. If limit ==
        'total_rainfall', this will be very close to
        median_total_rainfall_last_season. If 'total_time', it will diverge
        from this value.
        """
        return self._season_rf_limit


def Singer_orographic_rainfall(z_closest_node_to_center):
    """
    Return a set of curve weights for a provided z, assuming an orographic
    rule following that presented in Singer & Michaelides 2017 & Singer et
    al. 2018 and applicable specifically to Walnut Gulch.
    i.e., there are three orographic divisions, divided at 1350 m and
    1500 m.

    Parameters
    ----------
    z_closest_node_to_center : float
        The elevation of the node closest to the storm center.

    Returns
    -------
    wgts : length 11 list
        The weighting parameters to use in selecting a storm distribution
        curve.
    """
    if z_closest_node_to_center <= 1350:
        wgts = [
            0.0318,
            0.0759,
            0.0851,
            0.0941,
            0.0941,
            0.0941,
            0.0941,
            0.0941,
            0.1033,
            0.1121,
            0.1213,
        ]
    elif 1350 < z_closest_node_to_center <= 1500:
        wgts = [
            0.0478,
            0.0778,
            0.0869,
            0.0959,
            0.0959,
            0.0959,
            0.0959,
            0.0959,
            0.1051,
            0.1141,
            0.0888,
        ]
    elif z_closest_node_to_center > 1500:
        wgts = [
            0.0696,
            0.0786,
            0.0878,
            0.0968,
            0.0968,
            0.0968,
            0.0968,
            0.0968,
            0.1060,
            0.1149,
            0.0591,
        ]
    return wgts


if __name__ == "__main__":
    from matplotlib.pyplot import show

    nx = 17
    ny = 8
    dx = 1000.
    mg = RasterModelGrid((nx, ny), xy_spacing=dx)
    # closed_nodes = np.zeros((nx, ny), dtype=bool)
    # closed_nodes[:, :10] = True
    # closed_nodes[:, 30:] = True
    # closed_nodes[30:, :] = True
    # mg.status_at_node[closed_nodes.flatten()] = CLOSED_BOUNDARY
    # imshow_grid_at_node(mg, mg.status_at_node)
    # show()

    z = mg.add_zeros("node", "topographic__elevation")
    z += 1400.
    rain = SpatialPrecipitationDistribution(mg, number_of_years=1)
    count = 0
    total_t = 0.
    for dt, interval_t in rain.yield_storms(style="whole_year", limit="total_time"):
        count += 1
        total_t += dt + interval_t
        print(dt, interval_t)
        if count % 100 == 0:
            # imshow_grid_at_node(mg, rain.total_rainfall_this_year,
            #                     cmap='Blues')
            print("Season:", rain.current_season, "of yr", rain.current_year)
            print("Current storm:", count)
            # print('MEANS')
            # print(rain.total_rainfall_this_season[rain._opennodes].mean())
            # print(rain.total_rainfall_last_season[rain._opennodes].mean())
            # print(rain.total_rainfall_this_year[rain._opennodes].mean())
            # print(rain.total_rainfall_last_year[rain._opennodes].mean())
            # print('-----')
            # print('MEDIANS')
            # print(rain.median_total_rainfall_this_season)
            # print(rain.median_total_rainfall_last_season)
            # print(rain.median_total_rainfall_this_year)
            # print(rain.median_total_rainfall_last_year)
            # print('*****')
            show()
    print("Effective total years:")
    print(total_t / 24. / 365.)
    print("Storms simulated:")
    print(count)

    # mg = RasterModelGrid((100, 100), xy_spacing=500.)
    # # mg.status_at_node[closed_nodes.flatten()] = CLOSED_BOUNDARY
    # # imshow_grid_at_node(mg, mg.status_at_node)
    # # show()
    # z = mg.add_zeros('node', 'topographic__elevation')
    # z += 1000.
    # rain = SpatialPrecipitationDistribution(mg, number_of_years=2)
    # count = 0
    # total_storms = 0.
    # for storms_in_year in rain.yield_years():
    #     count += 1
    #     total_storms += storms_in_year
    #     print(storms_in_year)
    #     imshow_grid_at_node(mg, 'rainfall__total_depth_per_year', cmap='jet')
    #     show()
    #
    # _ = mg.at_node.pop('rainfall__total_depth_per_year')
    # _ = mg.at_node.pop('rainfall__flux')
    # rain = SpatialPrecipitationDistribution(mg, number_of_years=1)
    # for storms_in_season in rain.yield_seasons():
    #     print(storms_in_season)
    #     imshow_grid_at_node(mg, rain.total_rainfall_last_season, cmap='jet')
    #     show()
    #
    # for yr in range(30):
    #     print(rain.calc_annual_rainfall(style='whole_year'))
    #
    # from landlab import VoronoiDelaunayGrid
    #
    # x = np.random.rand(2000)*50000.
    # y = np.random.rand(2000)*50000.
    # vdg = VoronoiDelaunayGrid(x, y)
    # vdg.add_zeros('node', 'topographic__elevation')
    # rain = SpatialPrecipitationDistribution(vdg, number_of_years=2)
    # count = 0
    # total_storms = 0.
    # for storms_in_year in rain.yield_years(limit='total_rainfall'):
    #     count += 1
    #     total_storms += storms_in_year
    #     print(storms_in_year)
    #     imshow_grid_at_node(vdg, 'rainfall__total_depth_per_year',
    #     cmap='jet')
    #     print('Stochastically-set expected rainfall last season')
    #     print()
    #     print('Actual rainfall last season')
    #     print()
    #     show()



#! /usr/bin/env python

STANDARD_NAME = {
    "channel__bed_shear_stress": "channel_bottom_water__shear_stress",
    "channel__chi_index": None,
    "channel__depth": "channel_x-section__mean_of_depth",
    "channel__discharge": "channel_water_x-section__volume_flow_rate",
    "channel__steepness_index": "land_surface__steepness_index",
    "channel__width": "channel_x-section_top__width",
    "channel_sediment__relative_flux": "channel_water_sediment~suspended__time_max_normalized_volume_flux_ratio",
    "channel_sediment__volumetric_flux": "channel_water_sediment~suspended__volume_flux",  # (flux is area/time) or is it flow_rate?
    "channel_sediment__volumetric_transport_capacity": "channel_water_sediment~suspended__potential_volume_flow_rate",  # ?
    "depression__depth": "land_depression__max-fill_depth",
    "depression__outlet_node": "model_grid_node_land_depression_pour-point__id",  # necessary? model-specific? does anything need this?
    "drainage_area": "basin__total_contributing_area",  # model_grid_cell__total_contributing_area
    "flow__link_to_receiver_node": "model_grid_node_link~downstream__index",
    "flow__potential": None,
    "flow__receiver_node": None,
    "flow__sink_flag": None,  # model_*__flag boolean?
    "flow__upstream_node_order": None,
    "lithosphere__overlying_pressure_increment": "lithosphere_top_surface__increment_of_static_pressure",
    "lithosphere_surface__elevation_increment": "lithosphere_top_surface__increment_of_elevation",
    "lithosphere_surface__elevation_increment": "lithosphere_top_surface__increment_of_elevation",
    "plant__age": "plant__age",
    "plant__live_index": None,
    "radiation__incoming_shortwave_flux": "earth_surface_radiation~incoming~shortwave__energy_flux",
    "radiation__net_flux": "earth_surface_radiation~net~total__energy_flux",
    "radiation__net_longwave_flux": "earth_surface_radiation~net~longwave__energy_flux",
    "radiation__net_shortwave_flux": "earth_surface_radiation~net~shortwave__energy_flux",
    "radiation__ratio_to_flat_surface": "earth_surface_radiation~incoming~shortwave-to-flat_surface__energy_flux",
    "rainfall__daily_depth": "atmosphere_water__rainfall_volume_flux",
    "sediment_fill__depth": "land_surface_sediment__deposition_depth",
    "soil_moisture__cumulative_water_stress": "land_vegetation__time_integral_of_water_stress",
    "soil_moisture__initial_saturation_fraction": "soil_water__initial_saturated_volume_fraction",
    "soil_moisture__root_zone_leakage_rate": "soil_water_root-zone__volume_flux",
    "soil_moisture__saturation_fraction": "soil_water__saturated_volume_fraction",
    "soil_moisture__water_stress": "land_vegetation__water_stress",
    "surface__evapotranspiration_rate": "land_surface_water__evaptranspiration_volume_flux",
    "surface__potential_evapotranspiration_30day_mean": "land_surface_water__time_mean_of_potential_evaptranspiration_volume_flux",
    "surface__potential_evapotranspiration_rate": "land_surface_water__potential_evaptranspiration_volume_flux",
    "surface__runoff_rate": "land_surface_water__runoff_volume_flux",
    "surface_load__stress": "lithosphere_surface__normal_component_of_stress",
    "topographic__elevation": "land_surface__elevation",
    "topographic__gradient": "land_surface__slope",
    "topographic__slope": "land_surface__slope_angle",
    "topographic__steepest_slope": "model_grid_cell__max_of_d8_slope",
    "hillslope_sediment__unit_volume_flux": None,
    "vegetation__cover_fraction": "land_vegetation__area_fraction",
    "vegetation__dead_biomass": "land_vegetation~biomass~dead__volume_flux",
    "vegetation__dead_leaf_area_index": "land_vegetation~dead__leaf-area_index",
    "vegetation__live_biomass": "land_vegetation~biomass~live__volume_flux",
    "vegetation__live_leaf_area_index": "land_vegetation~live__leaf-area_index",
    "vegetation__plant_functional_type": None,
    "surface_water__depth": "land_surface_water__depth",
    "surface_water__discharge": "land_surface_water__volume_flow_rate",
    "water__unit_flux_in": "model_grid_cell_water~incoming__volume_flow_rate",
    "water_surface__gradient": "land_surface_water_surface__slope",
}


LANDLAB_NAME = dict((value, key) for key, value in STANDARD_NAME.items() if key)



# -*- coding: utf-8 -*-
"""
test_celllab_cts.py: Unit-test function for CellLabCTS.

Created on Thu Jul  9 08:20:06 2015

@author: gtucker
"""

from heapq import heappop, heappush

import numpy as np
from numpy.testing import assert_array_equal

from landlab import HexModelGrid, RasterModelGrid

# For dev
from landlab.ca.celllab_cts import Transition  # X, Event
from landlab.ca.celllab_cts import _RUN_NEW
from landlab.ca.hex_cts import HexCTS
from landlab.ca.oriented_hex_cts import OrientedHexCTS
from landlab.ca.oriented_raster_cts import OrientedRasterCTS
from landlab.ca.raster_cts import RasterCTS


def callback_function(ca, node1, node2, time_now):
    """
    This function is passed as an argument to a transition event, and then
    called automatically by CellLabCTSModel's do_event() method.
    """
    pass


#    elapsed_time = time_now - ca.last_update_time[node1]
#    if ca.node_state[node1]==1:
#        ca.prop_data[ca.propid[node1]]+=100*elapsed_time
#    elapsed_time = time_now - ca.last_update_time[node2]
#    if ca.node_state[node2]==1:
#        ca.prop_data[ca.propid[node2]]+=100*elapsed_time
#


def test_transition():
    """Test instantiation of Transition() object."""
    t = Transition(
        (0, 0, 0),
        (1, 1, 0),
        1.0,
        name="test",
        swap_properties=False,
        prop_update_fn=None,
    )
    assert t.from_state == (0, 0, 0)
    assert t.to_state == (1, 1, 0)
    assert t.rate == 1.0
    assert t.name == "test"
    assert t.swap_properties is False
    assert t.prop_update_fn is None


def test_raster_cts():
    """
    Tests instantiation of a RasterCTS and implementation of one transition,
    with a callback function.
    """

    # Set up a small grid with no events scheduled
    mg = RasterModelGrid(4, 4)
    mg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    node_state_grid = mg.add_ones("node", "node_state_map", dtype=int)
    node_state_grid[6] = 0
    ns_dict = {0: "black", 1: "white"}
    xn_list = []
    xn_list.append(Transition((1, 0, 0), (0, 1, 0), 0.1, "", True, callback_function))
    pd = mg.add_zeros("node", "property_data", dtype=int)
    pd[5] = 50
    ca = RasterCTS(
        mg, ns_dict, xn_list, node_state_grid, prop_data=pd, prop_reset_value=0
    )

    # Test the data structures
    assert ca.num_link_states == 4, "wrong number of link states"
    assert ca.prop_data[5] == 50, "error in property data"
    assert ca.num_node_states == 2, "error in num_node_states"
    assert ca.link_orientation[-1] == 0, "error in link orientation array"
    assert ca.link_state_dict[(1, 0, 0)] == 2, "error in link state dict"
    assert ca.n_xn[2] == 1, "error in n_xn"
    assert ca.node_pair[1] == (0, 1, 0), "error in cell_pair list"

    if _RUN_NEW:
        assert len(ca.priority_queue._queue) == 1, "event queue has wrong size"
        assert ca.next_trn_id.size == 24, "wrong size next_trn_id"
        assert ca.trn_id.shape == (4, 1), "wrong size for xn_to"
        assert ca.trn_id[2][0] == 0, "wrong value in xn_to"
        assert ca.trn_to[0] == 1, "wrong trn_to state"
        assert ca.trn_rate[0] == 0.1, "wrong trn rate"
        assert ca.trn_propswap[0] == 1, "wrong trn propswap"
        assert ca.trn_prop_update_fn == callback_function, "wrong prop upd"
    else:
        assert len(ca.event_queue) == 1, "event queue has wrong size"
        assert ca.xn_to.size == 4, "wrong size for xn_to"
        assert ca.xn_to.shape == (4, 1), "wrong size for xn_to"
        assert ca.xn_to[2][0] == 1, "wrong value in xn_to"
        assert ca.xn_rate[2][0] == 0.1, "error in transition rate array"

    # Manipulate the data in the event queue for testing:

    if _RUN_NEW:
        # pop the scheduled event off the queue
        (event_time, index, event_link) = ca.priority_queue.pop()
        assert (
            ca.priority_queue._queue == []
        ), "event queue should now be empty but is not"

        # engineer an event
        ca.priority_queue.push(8, 1.0)
        ca.next_update[8] = 1.0
        ca.next_trn_id[8] = 0

    else:
        # pop the scheduled event off the queue
        ev = heappop(ca.event_queue)
        assert ca.event_queue == [], "event queue should now be empty but is not"

        # engineer an event
        ev.time = 1.0
        ev.link = 8
        ev.xn_to = 1
        ev.propswap = True
        ev.prop_update_fn = callback_function
        ca.next_update[8] = 1.0

        # push it onto the event queue
        heappush(ca.event_queue, ev)

    # run the CA
    ca.run(2.0)

    # some more tests.
    # Is current time advancing correctly? (should only go to 1.0, not 2.0)
    # Did the two nodes (5 and 6) correctly exchange states?
    # Did the property ID and data arrays get updated? Note that the "propswap"
    # should switch propids between nodes 5 and 6, and the callback function
    # should increase the value of prop_data in the "swap" node from 50 to 150.
    assert ca.current_time == 1.0, "current time incorrect"
    assert ca.node_state[5] == 0, "error in node state 5"
    assert ca.node_state[6] == 1, "error in node state 6"
    # assert (ca.prop_data[ca.propid[6]]==150), 'error in prop swap'


def test_oriented_raster_cts():
    """Tests instantiation of an OrientedRasterCTS() object"""
    mg = RasterModelGrid(3, 3)
    nsd = {0: "oui", 1: "non"}
    xnlist = []
    xnlist.append(Transition((0, 1, 0), (1, 1, 0), 1.0, "hopping"))
    nsg = mg.add_zeros("node", "node_state_grid")
    orcts = OrientedRasterCTS(mg, nsd, xnlist, nsg)

    assert orcts.num_link_states == 8
    # assert_array_equal(orcts.link_orientation, [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    assert_array_equal(orcts.link_orientation, [0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0])


def test_hex_cts():
    """Tests instantiation of a HexCTS() object"""
    mg = HexModelGrid(3, 2, 1.0, orientation="vertical", reorient_links=True)
    nsd = {0: "zero", 1: "one"}
    xnlist = []
    xnlist.append(Transition((0, 1, 0), (1, 1, 0), 1.0, "transitioning"))
    nsg = mg.add_zeros("node", "node_state_grid")
    hcts = HexCTS(mg, nsd, xnlist, nsg)

    assert hcts.num_link_states == 4
    assert_array_equal(hcts.link_orientation, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])


def test_oriented_hex_cts():
    """Tests instantiation of an OrientedHexCTS() object"""
    mg = HexModelGrid(3, 2, 1.0, orientation="vertical", reorient_links=True)
    nsd = {0: "zero", 1: "one"}
    xnlist = []
    xnlist.append(Transition((0, 1, 0), (1, 1, 0), 1.0, "transitioning"))
    nsg = mg.add_zeros("node", "node_state_grid")
    ohcts = OrientedHexCTS(mg, nsd, xnlist, nsg)

    assert ohcts.num_link_states == 12
    # assert_array_equal(ohcts.link_orientation, [2, 1, 0, 0, 0, 2, 1, 0, 2, 1, 0])
    assert_array_equal(ohcts.link_orientation, [2, 0, 1, 0, 2, 0, 1, 0, 2, 0, 1])


def test_priority_queue():
    """Test import and use of priority queue."""
    from ..cfuncs import PriorityQueue

    # Create a priority queue
    pq = PriorityQueue()

    # push a bunch of events
    pq.push(2, 2.2)
    pq.push(5, 5.5)
    pq.push(0, 0.11)
    pq.push(4, 4.4)
    pq.push(1, 1.1)
    pq.push(3, 3.3)

    # pop a bunch of events
    (priority, index, item) = pq.pop()
    assert priority == 0.11, "incorrect priority in PQ test"
    assert index == 2, "incorrect index in PQ test"
    assert item == 0, "incorrect item in PQ test"

    (priority, index, item) = pq.pop()
    assert priority == 1.1, "incorrect priority in PQ test"
    assert index == 4, "incorrect index in PQ test"
    assert item == 1, "incorrect item in PQ test"

    (priority, index, item) = pq.pop()
    assert priority == 2.2, "incorrect priority in PQ test"
    assert index == 0, "incorrect index in PQ test"
    assert item == 2, "incorrect item in PQ test"

    (priority, index, item) = pq.pop()
    assert priority == 3.3, "incorrect priority in PQ test"
    assert index == 5, "incorrect index in PQ test"
    assert item == 3, "incorrect item in PQ test"

    (priority, index, item) = pq.pop()
    assert priority == 4.4, "incorrect priority in PQ test"
    assert index == 3, "incorrect index in PQ test"
    assert item == 4, "incorrect item in PQ test"

    (priority, index, item) = pq.pop()
    assert priority == 5.5, "incorrect priority in PQ test"
    assert index == 1, "incorrect index in PQ test"
    assert item == 5, "incorrect item in PQ test"


def test_run_oriented_raster():
    """Test running with a small grid, 2 states, 4 transition types."""

    # Create an OrientedRaster with a 3x5 raster grid. Test model has 2 node
    # states and 4 transition types.
    grid = RasterModelGrid((3, 5))
    nsd = {0: "zero", 1: "one"}
    trn_list = []
    trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
    trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
    trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
    trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
    ins = np.arange(15) % 2  # makes a checkerboard pattern
    cts = OrientedRasterCTS(grid, nsd, trn_list, ins)

    # Run to 1st transition, at ~0.12
    cts.run(0.15)
    assert_array_equal(cts.node_state, [0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0])

    # Run to 2nd transition, at ~0.19
    cts.run(0.2)
    assert_array_equal(cts.node_state, [0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0])

    # Run to 3rd transition, at ~0.265
    cts.run(0.27)
    assert_array_equal(cts.node_state, [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0])

    # Run to 4th transition, at ~0.276 (transition is ignored)
    cts.run(0.28)
    assert_array_equal(cts.node_state, [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0])

    # Run to 5th transition, at ~0.461 (ignored)
    cts.run(0.5)
    assert_array_equal(cts.node_state, [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0])

    # Run to 6th transition, at ~0.648
    cts.run(0.65)
    assert_array_equal(cts.node_state, [0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0])


def test_grain_hill_model():
    """Run a lattice-grain-based hillslope evolution model."""
    from .grain_hill import GrainHill

    params = {
        "number_of_node_rows": 10,
        "number_of_node_columns": 10,
        "report_interval": 5.0,
        "run_duration": 10.0,
        "output_interval": 1.0e5,
        "settling_rate": 220000000.0,
        "disturbance_rate": 0.01,
        "uplift_interval": 4000.0,
        "friction_coef": 1.0,
        "plot_interval": 1.0,
        "show_plots": False,
    }
    grid_size = (
        int(params["number_of_node_rows"]),
        int(params["number_of_node_columns"]),
    )
    grain_hill_model = GrainHill(grid_size, **params)
    grain_hill_model.run()

    # Now test
    assert_array_equal(
        grain_hill_model.grid.at_node["node_state"][:18],
        [8, 7, 7, 7, 7, 7, 7, 7, 7, 8, 0, 7, 7, 7, 7, 0, 7, 7],
    )

    # Try with an uplift step
    params["uplift_interval"] = 5.0
    params["run_duration"] = 15.0
    grain_hill_model = GrainHill(grid_size, **params)
    grain_hill_model.run()

    # Test
    assert_array_equal(
        grain_hill_model.grid.at_node["node_state"][20:38],
        [0, 7, 7, 7, 7, 0, 7, 7, 7, 0, 0, 0, 7, 7, 0, 0, 0, 7],
    )


def test_setup_transition_data():
    """Test the CellLabCTSModel setup_transition_data method."""
    grid = RasterModelGrid((3, 4))
    nsd = {0: "zero", 1: "one"}
    trn_list = []
    trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
    trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
    trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
    trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
    ins = np.arange(12) % 2
    cts = OrientedRasterCTS(grid, nsd, trn_list, ins)

    assert_array_equal(cts.n_trn, [0, 1, 1, 0, 0, 2, 0, 0])
    assert_array_equal(
        cts.trn_id, [[0, 0], [0, 0], [1, 0], [0, 0], [0, 0], [2, 3], [0, 0], [0, 0]]
    )
    assert_array_equal(cts.trn_to, [2, 1, 6, 7])
    assert_array_equal(cts.trn_rate, [1., 2., 3., 4.])


if __name__ == "__main__":
    test_transition()
    test_raster_cts()
    test_oriented_raster_cts()
    test_hex_cts()
    test_oriented_hex_cts()
    test_run_oriented_raster()
    test_grain_hill_model()



# -*- coding: utf-8 -*-
"""
colors.py

Created on Mon Jan 18 13:28:17 2016

@author: gtucker
"""

from matplotlib.colors import LinearSegmentedColormap


def water_colormap():
    """Return matplotlib colormap with 'water' theme."""
    cdict = {
        "red": ((0.0, 0.0, 169.0 / 255.0), (1.0, 38.0 / 255.0, 1.0)),
        "green": ((0.0, 0.0, 222.0 / 255.0), (1.0, 39.0 / 255.0, 1.0)),
        "blue": ((0.0, 0.0, 242.0 / 255.0), (1.0, 23.0 / 255.0, 1.0)),
    }
    return LinearSegmentedColormap("landlab_water", cdict)


def earth_colormap():
    """Return matplotlib colormap with 'earth' theme."""
    cdict = {
        "red": ((0.0, 0.0, 252.0 / 255.0), (1.0, 33.0 / 255.0, 1.0)),
        "green": ((0.0, 0.0, 237.0 / 255.0), (1.0, 38.0 / 255.0, 1.0)),
        "blue": ((0.0, 0.0, 179.0 / 255.0), (1.0, 24.0 / 255.0, 1.0)),
    }
    return LinearSegmentedColormap("landlab_earth", cdict)


def colormap(name):
    """Return named Landlab colormap as a matplotlib colormap.

    Parameters
    ----------
    name : str
        Name of colormap

    Currently available maps are:
        'water': black to light blue
        'earth': dark olive to light sand color
    """
    colormap_fns = {"water": water_colormap(), "earth": earth_colormap()}
    try:
        return colormap_fns[name]
    except KeyError:
        print('Warning: colormap "' + name + '" does not exist')
        return None



from .generate_overland_flow_Bates import OverlandFlowBates
from .generate_overland_flow_deAlmeida import OverlandFlow
from .generate_overland_flow_implicit_kinwave import KinwaveImplicitOverlandFlow
from .generate_overland_flow_kinwave import KinwaveOverlandFlowModel

__all__ = [
    "OverlandFlowBates",
    "OverlandFlow",
    "KinwaveImplicitOverlandFlow",
    "KinwaveOverlandFlowModel",
]



#! /usr/env/python

"""
flow_director_mfd.py: provides the component FlowDirectorMFD.

This components finds the steepest single-path steepest descent flow
directions. It is equivalent to D4 method in the special case of a raster grid
in that it does not consider diagonal links between nodes. For that capability,
use FlowDirectorD8.
"""

import numpy

from landlab import (
    BAD_INDEX_VALUE,
    FIXED_GRADIENT_BOUNDARY,
    FIXED_VALUE_BOUNDARY,
    VoronoiDelaunayGrid,
)
from landlab.components.flow_director import flow_direction_mfd
from landlab.components.flow_director.flow_director_to_many import _FlowDirectorToMany


class FlowDirectorMFD(_FlowDirectorToMany):

    """Multiple-path flow direction with or without out diagonals.

    Directs flow by the multiple flow direction method. Each node is assigned
    multiple flow directions, toward all of the N neighboring nodes that are
    lower than it. If none of the neighboring nodes are lower, the location is
    identified as a pit. Flow proportions can be calculated as proportional to
    slope or proportional to the square root of slope, which is the solution to
    a steady kinematic wave.

    Specifically, it stores as ModelGrid fields:

    -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
       there is no receiver: *'flow__receiver_node'*. This array is 2D, and is
       of dimension (number of nodes x max number of receivers).
    -  Node array of flow proportions: *'flow__receiver_proportions'*. This
       array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of links carrying flow:  *'flow__link_to_receiver_node'*.
       This array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of downhill slopes corresponding to each receiver.:
       *'topographic__steepest_slope'* This array is 2D, and is
       of dimension (number of nodes x max number of receivers). If the slope is
       uphill or flat, the value is assigned zero.
    -  Boolean node array of all local lows: *'flow__sink_flag'*
    -  Link array identifing if flow goes with (1) or against (-1) the link
       direction: *'flow_link_direction'*
    The primary method of this class is :func:`run_one_step`.

    Examples
    --------

    This method works for both raster and irregular grids. First we will look
    at a raster example, and then an irregular example.

    >>> import numpy as numpy
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowDirectorMFD
    >>> mg = RasterModelGrid((3,3), xy_spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field(
    ...     'topographic__elevation',
    ...     mg.node_x + mg.node_y,
    ...     at = 'node'
    ... )

    The MFD flow director can be uses for raster and irregular grids. For
    raster grids, use of diagonal links is specified with the keyword
    *diagonals* (default is False).

    >>> fd = FlowDirectorMFD(mg, 'topographic__elevation', diagonals = True)
    >>> fd.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    >>> fd.run_one_step()

    Unlike flow directors that only direct flow to one node, FlowDirectorMFD
    directs flow to all downstream nodes. It stores the receiver information
    is a (number of nodes x maximum number or receivers) shape field at nodes.

    >>> mg.at_node['flow__receiver_node']
    array([[ 0, -1, -1, -1, -1, -1, -1, -1],
           [ 1, -1, -1, -1, -1, -1, -1, -1],
           [ 2, -1, -1, -1, -1, -1, -1, -1],
           [ 3, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1,  1, -1, -1,  0, -1],
           [ 5, -1, -1, -1, -1, -1, -1, -1],
           [ 6, -1, -1, -1, -1, -1, -1, -1],
           [ 7, -1, -1, -1, -1, -1, -1, -1],
           [ 8, -1, -1, -1, -1, -1, -1, -1]])

    It also stores the proportions of flow going to each receiver, the link on
    which the flow moves in at node arrays, and the slope of each link.

    >>> mg.at_node['flow__receiver_proportions'] # doctest: +NORMALIZE_WHITESPACE
    array([[ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.41421356,  0.        ,
             0.        ,  0.58578644,  0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ]])
    >>> mg.at_node['flow__link_to_receiver_node']
    array([[-1, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1,  3, -1, -1, 12, -1],
           [-1, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1, -1]])
    >>> mg.at_node['topographic__steepest_slope'] # doctest: +NORMALIZE_WHITESPACE
    array([[ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  1.        ,  0.        ,
             0.        ,  1.41421356,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ,  0.        ,  0.        ]])

    Finally, FlowDirectorMFD identifies sinks, or local lows.

    >>> mg.at_node['flow__sink_flag']
    array([1, 1, 1, 1, 0, 1, 1, 1, 1], dtype=int8)

    The flow directors also have the ability to return the flow receiver nodes.
    For this example, we will turn the diagonals off. This is the default
    value.

    >>> fd = FlowDirectorMFD(mg, 'topographic__elevation')
    >>> fd.run_one_step()
    >>> receivers, proportions = fd.direct_flow()
    >>> receivers
    array([[ 0, -1, -1, -1],
           [ 1, -1, -1, -1],
           [ 2, -1, -1, -1],
           [ 3, -1, -1, -1],
           [-1, -1, -1,  1],
           [ 5, -1, -1, -1],
           [ 6, -1, -1, -1],
           [ 7, -1, -1, -1],
           [ 8, -1, -1, -1]])
    >>> proportions # doctest: +NORMALIZE_WHITESPACE
    array([[ 1.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.],
           [ 0.,  0.,  0.,  1.],
           [ 1.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.]])

    For each donor node (represented by each row) the proportions should sum to
    one.

    >>> proportions.sum(axis=1)
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])

    For the second example we will use a Hexagonal Model Grid, a special type
    of Voroni Grid that has regularly spaced hexagonal cells. FlowDirectorMFD
    has multiple ways to partition flow based on slope. The default method is
    based on the slope angle. A secondary methods is to partion based on the
    square root of slope. This represents the solution to a steady kinematic
    wave.

    >>> from landlab import HexModelGrid
    >>> mg = HexModelGrid(5,3)
    >>> _ = mg.add_field(
    ...     'topographic__elevation',
    ...     mg.node_x + numpy.round(mg.node_y),
    ...     at = 'node'
    ... )
    >>> fd = FlowDirectorMFD(
    ...      mg,
    ...      'topographic__elevation',
    ...      partition_method='square_root_of_slope'
    ...      )
    >>> fd.surface_values # doctest: +NORMALIZE_WHITESPACE
    array([ 1. ,  2. ,  3. ,
            1.5,  2.5,  3.5,  4.5,
            2. ,  3. ,  4. ,  5. ,  6. ,
            3.5,  4.5,  5.5,  6.5,
            4. ,  5. ,  6. ])
    >>> fd.run_one_step()
    >>> mg.at_node['flow__receiver_node']
    array([[ 0, -1, -1, -1, -1, -1],
           [ 1, -1, -1, -1, -1, -1],
           [ 2, -1, -1, -1, -1, -1],
           [ 3, -1, -1, -1, -1, -1],
           [-1, -1, -1,  3,  0,  1],
           [-1, -1, -1,  4,  1,  2],
           [ 6, -1, -1, -1, -1, -1],
           [ 7, -1, -1, -1, -1, -1],
           [-1, -1, -1,  7,  3,  4],
           [-1, -1, -1,  8,  4,  5],
           [-1, -1, -1,  9,  5,  6],
           [11, -1, -1, -1, -1, -1],
           [12, -1, -1, -1, -1, -1],
           [-1, -1, 16, 12,  8,  9],
           [-1, -1, 17, 13,  9, 10],
           [15, -1, -1, -1, -1, -1],
           [16, -1, -1, -1, -1, -1],
           [17, -1, -1, -1, -1, -1],
           [18, -1, -1, -1, -1, -1]])
    >>> mg.at_node['flow__receiver_proportions'] # doctest: +NORMALIZE_WHITESPACE
    array([[ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.34108138,  0.41773767,
             0.24118095],
           [ 0.        ,  0.        ,  0.        ,  0.34108138,  0.41773767,
             0.24118095],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 0.        ,  0.        ,  0.        ,  0.34108138,  0.41773767,
             0.24118095],
           [ 0.        ,  0.        ,  0.        ,  0.34108138,  0.41773767,
             0.24118095],
           [ 0.        ,  0.        ,  0.        ,  0.34108138,  0.41773767,
             0.24118095],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 0.        ,  0.        ,  0.19431571,  0.27480391,  0.33656468,
             0.19431571],
           [ 0.        ,  0.        ,  0.19431571,  0.27480391,  0.33656468,
             0.19431571],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ],
           [ 1.        ,  0.        ,  0.        ,  0.        ,  0.        ,
             0.        ]])
    >>> mg.at_node['flow__link_to_receiver_node']
    array([[-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1,  8,  3,  4],
           [-1, -1, -1,  9,  5,  6],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, 19, 12, 13],
           [-1, -1, -1, 20, 14, 15],
           [-1, -1, -1, 21, 16, 17],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, 35, 31, 25, 26],
           [-1, -1, 37, 32, 27, 28],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1]])
    >>> mg.at_node['topographic__steepest_slope'] # doctest: +NORMALIZE_WHITESPACE
    array([[ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  1. ,  1.5,  0.5],
           [ 0. ,  0. ,  0. ,  1. ,  1.5,  0.5],
           [ 0. ,  0. ,  1. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  1. ,  1.5,  0.5],
           [ 0. ,  0. ,  0. ,  1. ,  1.5,  0.5],
           [ 0. ,  0. ,  0. ,  1. ,  1.5,  0.5],
           [ 0. ,  1. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0.5,  0. ,  0. ],
           [ 0. ,  0. ,  0.5,  1. ,  1.5,  0.5],
           [ 0. ,  0. ,  0.5,  1. ,  1.5,  0.5],
           [ 0. ,  1. ,  1.5,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0.5,  0. ,  0. ,  0. ],
           [ 0. ,  0.5,  0. ,  0. ,  0. ,  0. ]])
    >>> mg.at_node['flow__sink_flag']
    array([1, 1, 1,
           1, 0, 0, 1,
           1, 0, 0, 0, 1,
           1, 0, 0, 1,
           1, 1, 1], dtype=int8)
    """

    _name = "FlowDirectorMFD"

    def __init__(self, grid, surface="topographic__elevation", **kwargs):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        surface : field name at node or array of length node, optional
            The surface to direct flow across, default is field at node:
            topographic__self.surface_valuesation.
        partition_method: string, optional
            Method for partitioning flow. Options include 'slope' (default) and
            'square_root_of_slope'.
        """
        # unpack kwargs:
        partition_method = kwargs.get("partition_method", "slope")
        diagonals = kwargs.get("diagonals", False)

        self.method = "MFD"
        super(FlowDirectorMFD, self).__init__(grid, surface)
        self._is_Voroni = isinstance(self._grid, VoronoiDelaunayGrid)
        if self._is_Voroni:
            diagonals = False
        self.updated_boundary_conditions()
        self.partition_method = partition_method
        self.diagonals = diagonals

        if self._is_Voroni is False and diagonals is False:
            self.max_receivers = 4
        if self._is_Voroni is False and diagonals is True:
            self.max_receivers = 8
        else:
            self.max_receivers = self._grid.adjacent_nodes_at_node.shape[1]

        # set the number of recievers, proportions, and receiver links with the
        # right size.
        self.receivers = grid.add_field(
            "flow__receiver_node",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=int),
            at="node",
            dtype=int,
            noclobber=False,
        )

        self.steepest_slope = grid.add_field(
            "topographic__steepest_slope",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=float),
            at="node",
            dtype=float,
            noclobber=False,
        )

        self.receiver_links = grid.add_field(
            "flow__link_to_receiver_node",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=int),
            at="node",
            dtype=int,
            noclobber=False,
        )

        self.proportions = grid.add_field(
            "flow__receiver_proportions",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=float),
            at="node",
            dtype=int,
            noclobber=False,
        )

    def updated_boundary_conditions(self):
        """Method to update FlowDirectorMFD when boundary conditions change.

        Call this if boundary conditions on the grid are updated after
        the component is instantiated.
        """
        self._active_links = self.grid.active_links
        self._activelink_tail = self.grid.node_at_link_tail[self.grid.active_links]
        self._activelink_head = self.grid.node_at_link_head[self.grid.active_links]

    def run_one_step(self):
        """Find flow directions and save to the model grid.

        run_one_step() checks for updated boundary conditions, calculates
        slopes on links, finds basself.surface_valuesel nodes based on the
        status at node, calculates flow directions, and saves results to the
        grid.

        An alternative to run_one_step() is direct_flow() which does the same
        things but also returns the receiver nodes not return values.
        """
        self.direct_flow()

    def direct_flow(self):
        """Find flow directions, save to the model grid, and return receivers.

        direct_flow() checks for updated boundary conditions, calculates
        slopes on links, finds basself.surface_valuesel nodes based on the status at node,
        calculates flow directions, saves results to the grid, and returns a
        at-node array  of receiver nodes. This array is stored in the grid at:
        grid['node']['flow__receiver_nodes']

        An alternative to direct_flow() is run_one_step() which does the same
        things but also returns a at-node array  of receiver nodes. This array
        is stored in the grid at:
        grid['node']['flow__receiver_nodes']
        """
        self._check_updated_bc()

        # step 1. Required inumpyuts for flow_directions_MFD
        # this is where diagonals are or are not included in
        # flow direction calculations

        # Option for no diagonals (default)
        if self.diagonals is False:
            neighbors_at_node = self.grid.adjacent_nodes_at_node
            links_at_node = self.grid.links_at_node
            active_link_dir_at_node = self.grid.active_link_dirs_at_node

            # this needs to be the gradient
            link_slope = self.grid.calc_grad_at_link(self.surface_values)

        # Option with diagonals.
        else:

            # need to create a list of diagonal links since it doesn't exist.
            diag_links = numpy.sort(numpy.unique(self.grid.d8s_at_node[:, 4:]))
            diag_links = diag_links[diag_links > 0]

            # get diagonal active links (though this actually includes ALL
            # active links)
            dal = self.grid.active_d8

            # calculate graidents across diagonals
            diag_grads = numpy.zeros(diag_links.shape)
            where_active_diag = dal >= diag_links.min()
            active_diags_inds = dal[where_active_diag] - diag_links.min()
            active_diag_grads = self.grid._calculate_gradients_at_d8_active_links(
                self.surface_values
            )
            diag_grads[active_diags_inds] = active_diag_grads[where_active_diag]

            # calculate gradients on orthogonal links
            ortho_grads = self.grid.calc_grad_at_link(self.surface_values)

            # concatenate the diagonal and orthogonal grid elements
            neighbors_at_node = numpy.hstack(
                (
                    self.grid.adjacent_nodes_at_node,
                    self.grid.diagonal_adjacent_nodes_at_node,
                )
            )
            active_link_dir_at_node = numpy.hstack(
                (
                    self.grid.active_link_dirs_at_node,
                    self.grid.active_diagonal_dirs_at_node,
                )
            )
            link_slope = numpy.hstack((ortho_grads, diag_grads))

            links_at_node = self.grid.d8s_at_node

        # Step 2. Find and save base level nodes.
        (baselevel_nodes,) = numpy.where(
            numpy.logical_or(
                self._grid.status_at_node == FIXED_VALUE_BOUNDARY,
                self._grid.status_at_node == FIXED_GRADIENT_BOUNDARY,
            )
        )

        # Calculate flow directions
        (
            self.receivers,
            self.proportions,
            slopes_to_receivers,
            steepest_slope,
            steepest_receiver,
            sink,
            receiver_links,
            steepest_link,
        ) = flow_direction_mfd.flow_directions_mfd(
            self.surface_values,
            neighbors_at_node,
            links_at_node,
            active_link_dir_at_node,
            link_slope,
            baselevel_nodes=baselevel_nodes,
            partition_method=self.partition_method,
        )

        # Save the four ouputs of this component.
        self._grid["node"]["flow__receiver_node"][:] = self.receivers
        self._grid["node"]["flow__receiver_proportions"][:] = self.proportions
        self._grid["node"]["topographic__steepest_slope"][:] = slopes_to_receivers
        self._grid["node"]["flow__link_to_receiver_node"][:] = receiver_links
        self._grid["node"]["flow__sink_flag"][:] = False
        self._grid["node"]["flow__sink_flag"][sink] = True

        return (self.receivers, self.proportions)


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



from landlab.field.scalar_data_fields import ScalarDataFields, FieldError
from landlab.field.grouped import ModelDataFields, GroupError, GroupSizeError
from landlab.field.field_mixin import ModelDataFieldsMixIn
from .graph_field import GraphFields

__all__ = [
    "ScalarDataFields",
    "ModelDataFields",
    "ModelDataFieldsMixIn",
    "FieldError",
    "GroupError",
    "GroupSizeError",
    "GraphFields",
]



import numpy as np
import pytest
from numpy.testing import assert_array_equal
from pytest import approx

from landlab import HexModelGrid, RadialModelGrid, RasterModelGrid


def test_xy_of_reference_default_is_zero():
    grid = RasterModelGrid((9, 5))
    assert grid.xy_of_reference == approx((0., 0.))


@pytest.mark.parametrize("to_iterable", [np.asarray, list, tuple])
def test_xy_of_reference_is_tuple(random_xy, to_iterable):
    grid = RasterModelGrid((9, 5), xy_of_reference=to_iterable(random_xy))
    assert isinstance(grid.xy_of_reference, tuple)
    assert grid.xy_of_reference == approx(random_xy)


def test_xy_of_reference_setter(random_xy):
    grid = RasterModelGrid((9, 5))
    grid.xy_of_reference = random_xy
    assert grid.xy_of_reference == approx(random_xy)


def test_move_reference_raster(random_xy):
    mg = RasterModelGrid((9, 5), xy_spacing=2.0, xy_of_lower_left=random_xy)
    assert mg.xy_of_lower_left == approx(random_xy)
    assert mg.x_of_node.min() == approx(random_xy[0])
    assert mg.y_of_node.min() == approx(random_xy[1])

    xy_of_new_lower_left = random_xy[0] * 10.0, random_xy[1] * 10.0

    mg.xy_of_lower_left = xy_of_new_lower_left
    assert mg.xy_of_lower_left == approx(xy_of_new_lower_left)
    assert mg.x_of_node.min() == approx(xy_of_new_lower_left[0])
    assert mg.y_of_node.min() == approx(xy_of_new_lower_left[1])


@pytest.mark.parametrize("to_iterable", [np.asarray, list, tuple])
def test_raster_lower_left_as_iterables(random_xy, to_iterable):
    expected = approx(tuple(random_xy))

    grid = RasterModelGrid((9, 5), xy_of_lower_left=to_iterable(random_xy))
    assert isinstance(grid.xy_of_lower_left, tuple)
    assert grid.xy_of_lower_left == expected


@pytest.mark.parametrize("to_iterable", [np.asarray, list, tuple])
def test_hex_lower_left_as_iterables(random_xy, to_iterable):
    expected = approx(tuple(random_xy))

    grid = HexModelGrid(
        9,
        5,
        xy_of_lower_left=to_iterable(random_xy),
        orientation="horizontal",
        shape="rect",
    )
    assert isinstance(grid.xy_of_lower_left, tuple)
    assert grid.xy_of_lower_left == expected


@pytest.mark.parametrize("to_iterable", [np.asarray, list, tuple])
def test_radial_center_as_iterables(random_xy, to_iterable):
    expected = approx(tuple(random_xy))

    grid = RadialModelGrid(num_shells=9, dr=10.0, xy_of_center=to_iterable(random_xy))
    assert isinstance(grid.xy_of_center, tuple)
    assert grid.xy_of_center == expected


@pytest.mark.parametrize("orientation", ["horizontal", "vertical"])
@pytest.mark.parametrize("shape", ["rect"])
@pytest.mark.parametrize("n_cols", [12, 11, 10, 9])
@pytest.mark.parametrize("n_rows", [12, 11, 10, 9])
def test_move_reference_hex(random_xy, n_rows, n_cols, shape, orientation):
    mg = HexModelGrid(
        n_rows,
        n_cols,
        dx=2.0,
        xy_of_lower_left=random_xy,
        orientation=orientation,
        shape=shape,
    )

    assert mg.xy_of_lower_left == random_xy
    assert mg.x_of_node.min() == random_xy[0]
    assert mg.y_of_node.min() == random_xy[1]

    mg.xy_of_lower_left = (30.0, 45.0)
    assert mg._xy_of_lower_left == (30.0, 45.0)
    assert mg.x_of_node.min() == approx(30.0)
    assert mg.y_of_node.min() == approx(45.0)


def test_move_reference_radial(random_xy):
    mg = RadialModelGrid(num_shells=9, dr=10.0, xy_of_center=random_xy)

    assert mg.xy_of_center == approx(random_xy)

    pre_move_llc = (mg.x_of_node.min(), mg.y_of_node.min())

    new_xy_of_center = (30.0, 45.0)
    mg.xy_of_center = new_xy_of_center
    assert mg.xy_of_center == approx(new_xy_of_center)

    post_move_llc = (mg.x_of_node.min(), mg.y_of_node.min())

    actual_dydx = (
        post_move_llc[0] - pre_move_llc[0],
        post_move_llc[1] - pre_move_llc[1],
    )
    known_dydx = (
        new_xy_of_center[0] - random_xy[0],
        new_xy_of_center[1] - random_xy[1],
    )

    assert known_dydx == approx(actual_dydx)


def test_radial_deprecate_origin_x():
    with pytest.deprecated_call():
        mg = RadialModelGrid(num_shells=1, dr=1.0, origin_x=10)
    assert mg._xy_of_center == (10.0, 0.0)
    pts, npts = mg._create_radial_points(1, 1, xy_of_center=mg._xy_of_center)
    assert pts[0, 0] == approx(mg._xy_of_center[0])
    assert pts[0, 1] == approx(mg._xy_of_center[1])


def test_radial_deprecate_origin_y():
    with pytest.deprecated_call():
        mg = RadialModelGrid(num_shells=1, dr=1.0, origin_y=10)
    assert mg._xy_of_center == (0.0, 10.0)
    pts, npts = mg._create_radial_points(1, 1, xy_of_center=mg._xy_of_center)
    assert pts[0, 0] == approx(mg._xy_of_center[0])
    assert pts[0, 1] == approx(mg._xy_of_center[1])


def test_raster_with_args_and_shape():
    with pytest.raises(ValueError):
        RasterModelGrid(3, 3, num_cols=3)


def test_raster_with_negative_shape():
    with pytest.raises(ValueError):
        RasterModelGrid(-2, 3)


def test_raise_deprecation_dx():
    with pytest.deprecated_call():
        mg = RasterModelGrid(3, 3, dx=3)
    X, Y = np.meshgrid([0.0, 3.0, 6.0], [0.0, 3.0, 6.0])
    assert_array_equal(mg.x_of_node, X.flatten())
    assert_array_equal(mg.y_of_node, Y.flatten())


def test_raise_deprecation_dxdx():
    with pytest.deprecated_call():
        mg = RasterModelGrid(3, 3, dx=(4, 5))
    X, Y = np.meshgrid([0.0, 5.0, 10.0], [0.0, 4.0, 8.0])
    assert_array_equal(mg.x_of_node, X.flatten())
    assert_array_equal(mg.y_of_node, Y.flatten())


def test_raise_deprecation_spacing():
    with pytest.deprecated_call():
        mg = RasterModelGrid(3, 3, spacing=(4, 5))
    X, Y = np.meshgrid([0.0, 5.0, 10.0], [0.0, 4.0, 8.0])
    assert_array_equal(mg.x_of_node, X.flatten())
    assert_array_equal(mg.y_of_node, Y.flatten())


def test_raise_deprecation_spacing_as_arg():
    with pytest.deprecated_call():
        mg = RasterModelGrid(3, 3, 3)
    X, Y = np.meshgrid([0.0, 3.0, 6.0], [0.0, 3.0, 6.0])
    assert_array_equal(mg.x_of_node, X.flatten())
    assert_array_equal(mg.y_of_node, Y.flatten())


def test_raise_deprecation_spacing2_as_arg():
    with pytest.deprecated_call():
        mg = RasterModelGrid(3, 3, (4, 5))
    X, Y = np.meshgrid([0.0, 5.0, 10.0], [0.0, 4.0, 8.0])
    assert_array_equal(mg.x_of_node, X.flatten())
    assert_array_equal(mg.y_of_node, Y.flatten())


def test_bad_shape_xy_spacing():
    with pytest.raises(ValueError):
        RasterModelGrid(3, 3, xy_spacing=(4, 5, 5))


def test_bad_type_xy_spacing():
    with pytest.raises(ValueError):
        RasterModelGrid(3, 3, xy_spacing="spam and eggs")


def test_deprecate_origin():
    with pytest.deprecated_call():
        mg = RasterModelGrid(3, 3, origin=(10, 13))
    X, Y = np.meshgrid([0.0, 1.0, 2.0], [0.0, 1.0, 2.0])
    assert_array_equal(mg.x_of_node, X.flatten() + 10.0)
    assert_array_equal(mg.y_of_node, Y.flatten() + 13.0)


def test_bad_origin():
    with pytest.raises(ValueError):
        RasterModelGrid(3, 3, xy_of_lower_left=(10, 13, 12))


def test_curent_vs_past_origin():
    with pytest.deprecated_call():
        mg1 = RasterModelGrid(3, 3, origin=(10, 13))
    mg2 = RasterModelGrid(3, 3, xy_of_lower_left=(10, 13))
    assert_array_equal(mg1.x_of_node, mg2.x_of_node)
    assert_array_equal(mg1.y_of_node, mg2.y_of_node)


def test_curent_vs_past_spacing():
    with pytest.deprecated_call():
        mg1 = RasterModelGrid(3, 3, spacing=(5, 4))
    mg2 = RasterModelGrid(3, 3, xy_spacing=(4, 5))
    assert_array_equal(mg1.x_of_node, mg2.x_of_node)
    assert_array_equal(mg1.y_of_node, mg2.y_of_node)



"""
Utility functions that operate on landlab grids.
------------------------------------------------

"""
import numpy as np


def resolve_values_on_active_links(grid, active_link_values):
    """Resolve active-link values into x and y directions.

    Takes a set of values defined on active links, and returns those values
    resolved into the x and y directions.  Two link arrays are returned:
    x, then y.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    active_link_values : ndarray
        Values on active links.

    Returns
    -------
    tuple of ndarray
        Values resolved into x-component and y-component.
    """
    link_lengths = grid.length_of_link[grid.active_links]
    return (
        np.multiply(
            (
                (
                    grid.node_x[grid._activelink_tonode]
                    - grid.node_x[grid._activelink_fromnode]
                )
                / link_lengths
            ),
            active_link_values,
        ),
        np.multiply(
            (
                (
                    grid.node_y[grid._activelink_tonode]
                    - grid.node_y[grid._activelink_fromnode]
                )
                / link_lengths
            ),
            active_link_values,
        ),
    )


def resolve_values_on_links(grid, link_values):
    """Resolve link values into x and y directions.

    Takes a set of values defined on active links, and returns those values
    resolved into the x and y directions.  Two link arrays are returned:
    x, then y.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    link_values : ndarray
        Values on links.

    Returns
    -------
    tuple of ndarray
        Values resolved into x-component and y-component.
    """
    return (
        np.multiply(
            (
                (
                    grid.node_x[grid.node_at_link_head]
                    - grid.node_x[grid.node_at_link_tail]
                )
                / grid.length_of_link
            ),
            link_values,
        ),
        np.multiply(
            (
                (
                    grid.node_y[grid.node_at_link_head]
                    - grid.node_y[grid.node_at_link_tail]
                )
                / grid.length_of_link
            ),
            link_values,
        ),
    )
